/*
		                  LCRZO
                             Network library
             Copyright(c) 1999,2000,2001, Laurent Constantin
                                  -----

  Main server    : http://www.laurentconstantin.com/
  Backup servers : http://laurentconstantin.free.fr/
                   http://go.to/laurentconstantin/
                   http://laurentconstantin.est-la.com/
                   http://laurentconstantin.ctw.net/
  [my current email address is on the web servers]

                                  -----
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details (http://www.gnu.org).

------------------------------------------------------------------------
*/

/* -- Note about comments and variable names --
   In versions 1.00 up to 2.08, lcrzo was only written in French (my
   mother tongue). Since version 3.00, lcrzo is written in English, but
   most of comments and variable names are still in French.
   In future versions, I'll try to translate everything into English,
   but as you may guess, that's a quite big work...
   So, for the moment, here are some word translation, if you want to
   be able to understand what's going on in the functions :
     afficher - print     ;  ajouter - add         ;  copie - copy
     dernier - last       ;  donnee - data         ;  ecrire - write
     entete - header      ;  envoye - sent         ;  ferme - close
     fichier - file       ;  lire - read           ;  masque - mask 
     nboct - byte number  ;  ouvre, ouvrir - open  ;  paq, paquet - packet
     plage - range        ;  premier - first       ;  recu - received
     reseau - network     ;  retour - return       ;  taille - size
     tunnel - pipe

*/

#include "lcrzo_incl.h"


/*---------------------------------------------------------------*/
int lcrzo_time_sleep_sec(lcrzo_uint32 sec)
{ 
  if (sec<lcrzo_global.time_sleep_threshold3/1000000)
  { /*utilise usleep*/
    usleep(sec*1000000);
  }
  else
  { /*utilise sleep*/
    sleep(sec);
  }

  errno=0; /*because Solaris sets errno to 4*/
  return(LCRZO_ERR_OK);
}

/*---------------------------------------------------------------*/
int lcrzo_time_sleep_usec(lcrzo_uint32 usec)
{ struct timeval tv1, tv2;
  int ret, oldprio;

  if (usec<lcrzo_global.time_sleep_threshold1)
  { ret=gettimeofday(&tv1, NULL);
    if (ret) return(LCRZO_ERR_FUGETTIMEOFDAY);
    tv2=tv1;
    /*on boucle en attendant l'heure*/
    while( (lcrzo_uint32)( (tv2.tv_sec-tv1.tv_sec)*1000000+
			   (tv2.tv_usec-tv1.tv_usec) ) < usec )
    { ret=gettimeofday(&tv2, NULL);
      if (ret) return(LCRZO_ERR_FUGETTIMEOFDAY);
    }
  }
  else if (usec<lcrzo_global.time_sleep_threshold2)
  { ret=gettimeofday(&tv1, NULL);
    if (ret) return(LCRZO_ERR_FUGETTIMEOFDAY);
    tv2=tv1;
    /*pour consommer moins de ressources*/
    oldprio=getpriority(PRIO_PROCESS, 0);
    setpriority(PRIO_PROCESS, 0, +20);
    /*on boucle en attendant l'heure*/
    while( (lcrzo_uint32)( (tv2.tv_sec-tv1.tv_sec)*1000000+
			   (tv2.tv_usec-tv1.tv_usec) ) < usec )
    { ret=gettimeofday(&tv2, NULL);
      if (ret) return(LCRZO_ERR_FUGETTIMEOFDAY);
    }
    setpriority(PRIO_PROCESS, 0, oldprio);
  }
  else if (usec<lcrzo_global.time_sleep_threshold3)
  { /*utilise usleep*/
    usleep(usec);
  }
  else
  { /*utilise sleep*/
    sleep(usec/1000000);
  }

  errno=0; /*because Solaris sets errno to 4*/
  return(LCRZO_ERR_OK);
}


/*---------------------------------------------------------------*/
int lcrzo_time_fprint(LCRZOFILE *pf,
		      const char *startstring,
		      const lcrzo_time *ptime,
		      const char *endstring)
{ 
  /*parameters verification*/
  if (pf==NULL) return(LCRZO_ERR_PANULLPTR);
  if (ptime==NULL) return(LCRZO_ERR_PANULLPTR);

  if (startstring!=NULL) fprintf(pf, "%s", startstring);
  fprintf(pf, "%lus:%06luus", ptime->sec, ptime->usec);
  if (endstring!=NULL) fprintf(pf, "%s", endstring);
  return(LCRZO_ERR_OK);
}


/*---------------------------------------------------------------*/
int lcrzo_time_init_currenttime(lcrzo_time *ptime)
{ struct timeval tv;
  int ret;

  /* parameters verification */
  if (ptime==NULL) return(LCRZO_ERR_PANULLPTR);

  ret=gettimeofday(&tv, NULL);
  if (ret)
  { return(LCRZO_ERR_FUGETTIMEOFDAY);
  }
  /* store values */
  ptime->sec=(lcrzo_uint32)tv.tv_sec;
  ptime->usec=(lcrzo_uint32)tv.tv_usec;
  return(LCRZO_ERR_OK);
}

/*---------------------------------------------------------------*/
int lcrzo_time_init_sec(lcrzo_uint32 sec,
			lcrzo_time *ptime)
{ 
  /* parameters verification */
  if (ptime==NULL) return(LCRZO_ERR_PANULLPTR);

  /* store values */
  ptime->sec=sec;
  ptime->usec=0;
  return(LCRZO_ERR_OK);
}

/*---------------------------------------------------------------*/
int lcrzo_time_init_usec(lcrzo_uint32 usec,
			 lcrzo_time *ptime)
{
  /* parameters verification */
  if (ptime==NULL) return(LCRZO_ERR_PANULLPTR);

  /* store values */
  if ( usec>999999 )
    ptime->sec=usec/1000000;
  else
    ptime->sec=0;

  ptime->usec = usec - 1000000*ptime->sec;
  return(LCRZO_ERR_OK);
}

/*---------------------------------------------------------------*/
int lcrzo_time_init_secusec(lcrzo_uint32 sec,
			    lcrzo_uint32 usec,
			    lcrzo_time *ptime)
{ 
  /* parameters verification */
  if (ptime==NULL) return(LCRZO_ERR_PANULLPTR);
  if ( usec>999999 ) return(LCRZO_ERR_BUTIMESECOVER);

  /* store values */
  ptime->sec=sec;
  ptime->usec=usec;
  return(LCRZO_ERR_OK);
}


/*---------------------------------------------------------------*/
int lcrzo_time_get_sec(const lcrzo_time *ptime,
		       lcrzo_uint32 *psec)
{ 
  /* parameters verification */
  if (ptime==NULL) return(LCRZO_ERR_PANULLPTR);

  /* store values */
  if (psec!=NULL) *psec=ptime->sec;
  return(LCRZO_ERR_OK);
}


/*---------------------------------------------------------------*/
int lcrzo_time_get_usec(const lcrzo_time *ptime,
			lcrzo_uint32 *pusec)
{ 
  /* parameters verification */
  if (ptime==NULL) return(LCRZO_ERR_PANULLPTR);

  if ( ptime->sec>4294 ) return(LCRZO_ERR_BUTIMEUSECOVER);
  if ( ptime->sec==4294 && ptime->usec>967296 )
    return(LCRZO_ERR_BUTIMEUSECOVER);

  /* store values */
  if (pusec!=NULL) *pusec= 1000000*ptime->sec + ptime->usec;
  return(LCRZO_ERR_OK);
}

/*---------------------------------------------------------------*/
int lcrzo_time_get_secusec(const lcrzo_time *ptime,
			   lcrzo_uint32 *psec,
			   lcrzo_uint32 *pusec)
{ 
  /* parameters verification */
  if (ptime==NULL) return(LCRZO_ERR_PANULLPTR);

  /* store values */
  if (psec!=NULL) *psec=ptime->sec;
  if (pusec!=NULL) *pusec=ptime->usec;
  return(LCRZO_ERR_OK);
}


/*---------------------------------------------------------------*/
int lcrzo_time_plus_sec(const lcrzo_time *ptime,
			lcrzo_uint32 sec,
			lcrzo_time *presulttime)
{ lcrzo_time tmptime;

  lcrzo_er(lcrzo_time_init_sec(sec, &tmptime));
  lcrzo_er(lcrzo_time_plus_time(ptime, &tmptime, presulttime));
  return(LCRZO_ERR_OK);
}

/*---------------------------------------------------------------*/
int lcrzo_time_plus_usec(const lcrzo_time *ptime,
			 lcrzo_uint32 usec,
			 lcrzo_time *presulttime)
{ lcrzo_time tmptime;

  lcrzo_er(lcrzo_time_init_usec(usec, &tmptime));
  lcrzo_er(lcrzo_time_plus_time(ptime, &tmptime, presulttime));
  return(LCRZO_ERR_OK);
}

/*---------------------------------------------------------------*/
int lcrzo_time_plus_time(const lcrzo_time *ptime1,
			 const lcrzo_time *ptime2,
			 lcrzo_time *presulttime)
{ lcrzo_time resulttime;
  lcrzo_uint32 usecsum, secsum;

  if (ptime1==NULL) return(LCRZO_ERR_PANULLPTR);
  if (ptime2==NULL) return(LCRZO_ERR_PANULLPTR);

  /* compute microseconds */
  usecsum=ptime1->usec+ptime2->usec;
  if ( usecsum>999999 )
    resulttime.sec = usecsum/1000000;
  else
    resulttime.sec=0;
  resulttime.usec = usecsum - 1000000*resulttime.sec;

  /* compute seconds */
  secsum=ptime1->sec+resulttime.sec;
  if ( secsum<ptime1->sec || secsum<resulttime.sec )
    return(LCRZO_ERR_BUTIMESECOVER);
  resulttime.sec=secsum;
  secsum=ptime2->sec+resulttime.sec;
  if ( secsum<ptime2->sec || secsum<resulttime.sec )
    return(LCRZO_ERR_BUTIMESECOVER);
  resulttime.sec=secsum;

  /* set result */
  if (presulttime!=NULL)
  { presulttime->sec=resulttime.sec;
    presulttime->usec=resulttime.usec;
  }

  return(LCRZO_ERR_OK);
}


/*---------------------------------------------------------------*/
int lcrzo_time_minus_sec(const lcrzo_time *ptime,
			 lcrzo_uint32 sec,
			 lcrzo_time *presulttime)
{ lcrzo_time tmptime;

  lcrzo_er(lcrzo_time_init_sec(sec, &tmptime));
  lcrzo_er(lcrzo_time_minus_time(ptime, &tmptime, presulttime));
  return(LCRZO_ERR_OK);
}

/*---------------------------------------------------------------*/
int lcrzo_time_minus_usec(const lcrzo_time *ptime,
			  lcrzo_uint32 usec,
			  lcrzo_time *presulttime)
{ lcrzo_time tmptime;

  lcrzo_er(lcrzo_time_init_usec(usec, &tmptime));
  lcrzo_er(lcrzo_time_minus_time(ptime, &tmptime, presulttime));
  return(LCRZO_ERR_OK);
}

/*---------------------------------------------------------------*/
int lcrzo_time_minus_time(const lcrzo_time *ptime1,
			  const lcrzo_time *ptime2,
			  lcrzo_time *presulttime)
{ lcrzo_uint32 usec, sec;

  if (ptime1==NULL) return(LCRZO_ERR_PANULLPTR);
  if (ptime2==NULL) return(LCRZO_ERR_PANULLPTR);

  /* negative difference */
  if ( ptime1->sec < ptime2->sec ) return(LCRZO_ERR_BUTIMEDIFFNEG);
  if ( ptime1->sec == ptime2->sec && ptime1->usec < ptime2->usec )
    return(LCRZO_ERR_BUTIMEDIFFNEG);

  /* compute */
  if ( ptime1->usec >= ptime2->usec )
  { sec =  ptime1->sec  - ptime2->sec;
    usec = ptime1->usec - ptime2->usec;
  }
  else
  { sec =  ptime1->sec  - ptime2->sec  -1 ;
    usec = ptime1->usec - ptime2->usec +1000000;
  }

  /* set result */
  if (presulttime!=NULL)
  { presulttime->sec=sec;
    presulttime->usec=usec;
  }

  return(LCRZO_ERR_OK);
}


/*---------------------------------------------------------------*/
int lcrzo_time_diffcurrent_sec(const lcrzo_time *ptime,
			       lcrzo_uint32 *psec)
{ lcrzo_time currenttime, tmptime;

  lcrzo_er(lcrzo_time_init_currenttime(&currenttime));
  lcrzo_er(lcrzo_time_minus_time(&currenttime, ptime, &tmptime));
  lcrzo_er(lcrzo_time_get_sec(&tmptime, psec));
  return(LCRZO_ERR_OK);
}

/*---------------------------------------------------------------*/
int lcrzo_time_diffcurrent_usec(const lcrzo_time *ptime,
				lcrzo_uint32 *pusec)
{ lcrzo_time currenttime, tmptime;

  lcrzo_er(lcrzo_time_init_currenttime(&currenttime));
  lcrzo_er(lcrzo_time_minus_time(&currenttime, ptime, &tmptime));
  lcrzo_er(lcrzo_time_get_usec(&tmptime, pusec));
  return(LCRZO_ERR_OK);
}

/*---------------------------------------------------------------*/
int lcrzo_time_diffcurrent_secusec(const lcrzo_time *ptime,
				   lcrzo_uint32 *psec,
				   lcrzo_uint32 *pusec)
{ lcrzo_time currenttime, tmptime;

  lcrzo_er(lcrzo_time_init_currenttime(&currenttime));
  lcrzo_er(lcrzo_time_minus_time(&currenttime, ptime, &tmptime));
  lcrzo_er(lcrzo_time_get_secusec(&tmptime, psec, pusec));
  return(LCRZO_ERR_OK);
}

/*---------------------------------------------------------------*/
int lcrzo_time_diffcurrent_time(const lcrzo_time *ptime,
				lcrzo_time *pdifftime)
{ lcrzo_time currenttime;

  lcrzo_er(lcrzo_time_init_currenttime(&currenttime));
  lcrzo_er(lcrzo_time_minus_time(&currenttime, ptime, pdifftime));
  return(LCRZO_ERR_OK);
}
