/*en    Laurent Constantin's network library (lcrzo)
                 lcrzo_time module

  Functions herein allow to measure time.
*/
/*fr    Bibliotheque reseau de Laurent Constantin (lcrzo)
                 Module lcrzo_time

  Les fonctions presentes dans ce module assurent la gestion
du temps.
*/

/*---------------------------------------------------------------*/
/*en definition of a lcrzo_time. You should not use its
   fields directly. */
/*fr definition d'un lcrzo_time. Vous ne devriez pas utiliser
   directement ses champs. */
typedef struct
{ lcrzo_uint32 sec;
  lcrzo_uint32 usec;  /* cannot be over 999999 */
  /* time in micro-seconds=sec*1000000+usec */
} lcrzo_time;


/*---------------------------------------------------------------*/
/*en Name : lcrzo_time_sleep_xsec
   Description :
     Wait for a number a microseconds or seconds.
   Input parameter(s) :
     sec : number of seconds to wait
     usec : number of microseconds to wait
   Input/output parameter(s) :
   Output parameter(s) :
   Normal return values :
     LCRZO_ERR_OK : ok
*/
/*fr Nom : lcrzo_time_sleep_xsec
   Description :
     Attend usec microsecondes, ou sec secondes
   Parametre(s) d'entree :
     sec : nombre de secondes a attendre
     usec : nombre de microsecondes a attendre
   Parametre(s) d'entree-sortie :
   Parametre(s) de sortie :
   Valeurs de retour normales :
     LCRZO_ERR_OK : ok
*/
int lcrzo_time_sleep_sec(lcrzo_uint32 sec);
int lcrzo_time_sleep_usec(lcrzo_uint32 usec);

/*---------------------------------------------------------------*/
/*en Name : lcrzo_time_fprint
   Description :
     Print a lcrzo_time.
   Input parameter(s) :
     *pf : FILE where the function prints
     startstring : string to print before the lcrzo_time
     *ptime : lcrzo_time to print
     endstring : string to print after the lcrzo_time
   Input/output parameter(s) :
   Output parameter(s) :
   Normal return values :
     LCRZO_ERR_OK : ok
*/
/*fr Nom : lcrzo_time_fprint
   Description :
     Affiche un lcrzo_time.
   Parametre(s) d'entree :
     *pf : FILE ou la fonction doit ecrire
     startstring : chaine a ecrire avant le lcrzo_time
     time : lcrzo_time a afficher
     endstring : chaine a ecrire apres le lcrzo_time
   Parametre(s) d'entree-sortie :
   Parametre(s) de sortie :
   Valeurs de retour normales :
     LCRZO_ERR_OK : ok
*/
int lcrzo_time_fprint(LCRZOFILE *pf,
		      const char *startstring,
		      const lcrzo_time *ptime,
		      const char *endstring);

/*en idem, but on stdout */
/*fr idem, mais sur stdout */
#define lcrzo_time_print(s,t,e) lcrzo_time_fprint(stdout, s,t,e)

/*---------------------------------------------------------------*/
/*en Name : lcrzo_time_init_currenttime
   Description :
     Initialize a lcrzo_time with the current time.
   Input parameter(s) :
   Input/output parameter(s) :
   Output parameter(s) :
     *ptime : lcrzo_time set to the current time
   Normal return values :
     LCRZO_ERR_OK : ok
*/
/*fr Nom : lcrzo_time_init_currenttime
   Description :
     Initialise un lcrzo_time avec l'heure courante.
   Parametre(s) d'entree :
   Parametre(s) d'entree-sortie :
   Parametre(s) de sortie :
     *ptime : lcrzo_time affecte a l'heure courante
   Valeurs de retour normales :
     LCRZO_ERR_OK : ok
*/
int lcrzo_time_init_currenttime(lcrzo_time *ptime);

/*en Name : lcrzo_time_init_sec
   Description :
     Initialize a lcrzo_time with a number of seconds.
   Input parameter(s) :
     sec : number of seconds
   Input/output parameter(s) :
   Output parameter(s) :
     *ptime : lcrzo_time set
   Normal return values :
     LCRZO_ERR_OK : ok
*/
/*fr Nom : lcrzo_time_init_sec
   Description :
     Initialise un lcrzo_time avec un nombre de secondes.
   Parametre(s) d'entree :
     sec : nombre de secondes
   Parametre(s) d'entree-sortie :
   Parametre(s) de sortie :
     *ptime : lcrzo_time affecte
   Valeurs de retour normales :
     LCRZO_ERR_OK : ok
*/
int lcrzo_time_init_sec(lcrzo_uint32 sec,
			lcrzo_time *ptime);

/*en Name : lcrzo_time_init_usec
   Description :
     Initialize a lcrzo_time with a number of micro-seconds.
   Input parameter(s) :
     usec : number of micro-seconds
   Input/output parameter(s) :
   Output parameter(s) :
     *ptime : lcrzo_time set
   Normal return values :
     LCRZO_ERR_OK : ok
*/
/*fr Nom : lcrzo_time_init_usec
   Description :
     Initialise un lcrzo_time avec un nombre de micro-secondes.
   Parametre(s) d'entree :
     usec : nombre de micro-secondes
   Parametre(s) d'entree-sortie :
   Parametre(s) de sortie :
     *ptime : lcrzo_time affecte
   Valeurs de retour normales :
     LCRZO_ERR_OK : ok
*/
int lcrzo_time_init_usec(lcrzo_uint32 usec,
			 lcrzo_time *ptime);

/*en Name : lcrzo_time_init_secusec
   Description :
     Initialize a lcrzo_time with a number of seconds.
   Input parameter(s) :
     sec : number of seconds
     usec : number of micro-seconds
   Input/output parameter(s) :
   Output parameter(s) :
     *ptime : lcrzo_time set
   Normal return values :
     LCRZO_ERR_OK : ok
*/
/*fr Nom : lcrzo_time_init_secusec
   Description :
     Initialise un lcrzo_time avec un nombre de secondes.
   Parametre(s) d'entree :
     sec : nombre de secondes
     usec : nombre de micro-secondes
   Parametre(s) d'entree-sortie :
   Parametre(s) de sortie :
     *ptime : lcrzo_time affecte
   Valeurs de retour normales :
     LCRZO_ERR_OK : ok
*/
int lcrzo_time_init_secusec(lcrzo_uint32 sec,
			    lcrzo_uint32 usec,
			    lcrzo_time *ptime);

/*en get seconds and microseconds stored in a lcrzo_time */
/*fr obtient les secondes et microsecondes stockees dans
  un lcrzo_time */
int lcrzo_time_get_sec(const lcrzo_time *ptime,
		       lcrzo_uint32 *psec);
/* note : this function gets lcrzo_time.usec+1000000*lcrzo_time.sec,
  but not the value of the field lcrzo_time.usec. Use
  lcrzo_time_get_secusec to obtain field's values */
int lcrzo_time_get_usec(const lcrzo_time *ptime,
			lcrzo_uint32 *pusec);
int lcrzo_time_get_secusec(const lcrzo_time *ptime,
			   lcrzo_uint32 *psec,
			   lcrzo_uint32 *pusec);

/*en add seconds/microseconds/time to a lcrzo_time */
/*fr ajoute des secondes/microsecondes/time a un lcrzo_time */
int lcrzo_time_plus_sec(const lcrzo_time *ptime,
			lcrzo_uint32 sec,
			lcrzo_time *presulttime);
int lcrzo_time_plus_usec(const lcrzo_time *ptime,
			 lcrzo_uint32 usec,
			 lcrzo_time *presulttime);
/* *ppresulttime = *ptime1 + *ptime2 */
int lcrzo_time_plus_time(const lcrzo_time *ptime1,
			 const lcrzo_time *ptime2,
			 lcrzo_time *presulttime);

/*en substract seconds/microseconds/time to a lcrzo_time */
/*fr retranche des secondes/microsecondes/time a un lcrzo_time */
int lcrzo_time_minus_sec(const lcrzo_time *ptime,
			 lcrzo_uint32 sec,
			 lcrzo_time *presulttime);
int lcrzo_time_minus_usec(const lcrzo_time *ptime,
			  lcrzo_uint32 usec,
			  lcrzo_time *presulttime);
/* *ppresulttime = *ptime1 - *ptime2 */
int lcrzo_time_minus_time(const lcrzo_time *ptime1,
			  const lcrzo_time *ptime2,
			  lcrzo_time *presulttime);

/*---------------------------------------------------------------*/
/*en Name : lcrzo_time_diffcurrent_sec
   Description :
     Compare a lcrzo_time to the current time.
   Input parameter(s) :
     *ptime : lcrzo_time to compare to the current time
   Input/output parameter(s) :
   Output parameter(s) :
     *psec : number of seconds between the two times
   Normal return values :
     LCRZO_ERR_OK : ok
*/
/*fr Nom : lcrzo_time_diffcurrent_sec
   Description :
     Compare un lcrzo_time a l'heure courante.
   Parametre(s) d'entree :
     *ptime : lcrzo_time a comparer a l'heure courante
   Parametre(s) d'entree-sortie :
   Parametre(s) de sortie :
     *psec : nombre de secondes entre les deux heures
   Valeurs de retour normales :
     LCRZO_ERR_OK : ok
*/
int lcrzo_time_diffcurrent_sec(const lcrzo_time *ptime,
			       lcrzo_uint32 *psec);

/*en Name : lcrzo_time_diffcurrent_usec
   Description :
     Compare a lcrzo_time to the current time.
   Input parameter(s) :
     *ptime : lcrzo_time to compare to the current time
   Input/output parameter(s) :
   Output parameter(s) :
     *pusec : number of microseconds between the two times
   Normal return values :
     LCRZO_ERR_OK : ok
*/
/*fr Nom : lcrzo_time_diffcurrent_usec
   Description :
     Compare un lcrzo_time a l'heure courante.
   Parametre(s) d'entree :
     *ptime : lcrzo_time a comparer a l'heure courante
   Parametre(s) d'entree-sortie :
   Parametre(s) de sortie :
     *pusec : nombre de microsecondes entre les deux heures
   Valeurs de retour normales :
     LCRZO_ERR_OK : ok
*/
int lcrzo_time_diffcurrent_usec(const lcrzo_time *ptime,
				lcrzo_uint32 *pusec);

/*en Name : lcrzo_time_diffcurrent_secusec
   Description :
     Compare a lcrzo_time to the current time.
   Input parameter(s) :
     *ptime : lcrzo_time to compare to the current time
   Input/output parameter(s) :
   Output parameter(s) :
     *psec : number of seconds between the two times
     *pusec : number of microseconds between the two times
   Normal return values :
     LCRZO_ERR_OK : ok
*/
/*fr Nom : lcrzo_time_diffcurrent_secusec
   Description :
     Compare un lcrzo_time a l'heure courante.
   Parametre(s) d'entree :
     *ptime : lcrzo_time a comparer a l'heure courante
   Parametre(s) d'entree-sortie :
   Parametre(s) de sortie :
     *psec : nombre de secondes entre les deux heures
     *pusec : nombre de microsecondes entre les deux heures
   Valeurs de retour normales :
     LCRZO_ERR_OK : ok
*/
int lcrzo_time_diffcurrent_secusec(const lcrzo_time *ptime,
				   lcrzo_uint32 *psec,
				   lcrzo_uint32 *pusec);

/*en Name : lcrzo_time_diffcurrent_time
   Description :
     Compare a lcrzo_time to the current time.
   Input parameter(s) :
     *ptime : lcrzo_time to compare to the current time
   Input/output parameter(s) :
   Output parameter(s) :
     *pdifftime : lcrzo_time containing the difference
   Normal return values :
     LCRZO_ERR_OK : ok
*/
/*fr Nom : lcrzo_time_diffcurrent_time
   Description :
     Compare un lcrzo_time a l'heure courante.
   Parametre(s) d'entree :
     *ptime : lcrzo_time a comparer a l'heure courante
   Parametre(s) d'entree-sortie :
   Parametre(s) de sortie :
     *pdifftime : lcrzo_time contenant la difference
   Valeurs de retour normales :
     LCRZO_ERR_OK : ok
*/
int lcrzo_time_diffcurrent_time(const lcrzo_time *ptime,
				lcrzo_time *pdifftime);
