
/* include the library headers */
#include <lcrzo.h>
#include "lcrzoex.h"
#include "shared/shared_misc.h"

/*-------------------------------------------------------------*/
/* set information about this example */
lcrzoex_infos lcrzoex_000246=
{ /*reference*/     "lcrzoex", 246,
  /*version*/       1, 0,
  /*title*/         "icmp traceroute (spoofed)",
  /*french title*/  "traceroute icmp (spoofe)",
  /*description*/   "",
  /*french desc.*/  "",
  /*usage*/         "device srceth routereth srcip dstip maxttl",
  /*french usage*/  "device ethsrc ethrouteur ipsrc ipdst ttlmax",
  /*usage example*/ "eth0 aa:bb:cc:dd:ee:ff 11:22:33:44:55:66 1.2.3.4 5.6.7.78 30",
  /*fr. usage ex.*/ "eth0 aa:bb:cc:dd:ee:ff 11:22:33:44:55:66 1.2.3.4 5.6.7.78 30",
  /*author*/        "Laurent"
};

/*-------------------------------------------------------------*/
/* function declarations */
int lcrzoex_000246_main(int argc, char *argv[]);
void lcrzoex_000246_sig_received1(int i);
void lcrzoex_000246_sig_received2(int i);
int lcrzoex_000246_traceroute(lcrzo_device device,
			      lcrzo_etha clientetha,
			      lcrzo_etha routeretha,
			      lcrzo_ipl clientipl, 
			      lcrzo_ipl serveripl, 
			      lcrzo_uint8 maxttl);
int lcrzoex_000246_sendecho(lcrzo_device device,
			    lcrzo_etha clientetha,
			    lcrzo_etha routeretha,
			    lcrzo_ipl clientipl, 
			    lcrzo_ipl serveripl, 
			    lcrzo_uint16 startingid,
			    lcrzo_uint8 maxttl);
int lcrzoex_000246_print(lcrzo_device device,
			 lcrzo_etha clientetha,
			 lcrzo_etha routeretha,
			 lcrzo_ipl clientipl, 
			 lcrzo_ipl serveripl, 
			 lcrzo_uint16 startingid,
			 lcrzo_uint8 maxttl);
int lcrzoex_000246_printroutei(lcrzo_etha clientetha,
			       lcrzo_etha routeretha,
			       lcrzo_ipl clientipl, 
			       lcrzo_ipl serveripl, 
			       lcrzo_uint16 startingid,
			       lcrzo_uint8 maxttl,
			       lcrzo_data sniffdata,
			       lcrzo_int32 sniffdatasize);
int lcrzoex_000246_printrouter(lcrzo_etha clientetha,
			       lcrzo_etha routeretha,
			       lcrzo_ipl clientipl, 
			       lcrzo_ipl serveripl, 
			       lcrzo_uint16 startingid,
			       lcrzo_data sniffdata,
			       lcrzo_int32 sniffdatasize);

/*-------------------------------------------------------------*/
/* global variables */
int lcrzoex_000246_pid;
lcrzo_spoof lcrzoex_000246_spoof;
lcrzo_sniff lcrzoex_000246_sniff;

/*-------------------------------------------------------------*/
/* main function : you can rename it to "main", and then this file
   can be compiled alone. */
int lcrzoex_000246_main(int argc, char *argv[])
{ lcrzo_device device;
  lcrzo_etha clientetha, routeretha;
  lcrzo_ipl clientipl, serveripl;
  lcrzo_uint8 maxttl;

  if ( argc<7 )
  { /* print usage */
    lcrzo_epr(lcrzoex_infos_print(lcrzoex_000246, argv[0]));
    return(LCRZO_ERR_BULCRZOEXBADUSAGE);
  }
  lcrzo_epr(lcrzo_device_init(argv[1], device));
  lcrzo_epr(lcrzo_etha_init_eths(argv[2], clientetha));
  lcrzo_epr(lcrzo_etha_init_eths(argv[3], routeretha));
  lcrzo_epr(lcrzo_ipl_init_hs(argv[4], &clientipl));
  lcrzo_epr(lcrzo_ipl_init_hs(argv[5], &serveripl));
  maxttl=(lcrzo_uint8)atoi(argv[6]); if (maxttl==0) maxttl=1;

  /* call the main function */
  lcrzo_epr(lcrzoex_000246_traceroute(device, clientetha, routeretha,
				      clientipl, serveripl,
				      maxttl));

  return (LCRZO_ERR_OK);
}

/*-------------------------------------------------------------*/
int lcrzoex_000246_traceroute(lcrzo_device device,
			      lcrzo_etha clientetha,
			      lcrzo_etha routeretha,
			      lcrzo_ipl clientipl, 
			      lcrzo_ipl serveripl, 
			      lcrzo_uint8 maxttl)
{ lcrzo_uint16 startingid;

  /* determine starting values */
  lcrzo_epr(lcrzo_uint16_rand(5000, 30000, &startingid));

  lcrzoex_000246_pid=fork();
  if (lcrzoex_000246_pid<0) return(LCRZO_ERR_FUFORK);
  if (lcrzoex_000246_pid==0)
  { /* this process sends an echo request */
    /* in case we receive a signal */
    signal(SIGINT, lcrzoex_000246_sig_received1);
    lcrzo_epr(lcrzoex_000246_sendecho(device, clientetha, routeretha,
				      clientipl, serveripl, startingid, 
				      maxttl));
  }
  else
  { /* this process sniff and prints ip addresses */
    /* in case we receive a signal */
    signal(SIGINT, lcrzoex_000246_sig_received2);
    lcrzo_epr(lcrzoex_000246_print(device, clientetha, routeretha,
				   clientipl, serveripl, startingid, 
				   maxttl));
  }
  return (LCRZO_ERR_OK);
}

/*-------------------------------------------------------------*/
void lcrzoex_000246_sig_received1(int i)
{ int language;

  lcrzo_global_get_language(&language);
  if ( language==LCRZO_GLOBAL_LANG_FR )
    printf("Fermeture du spoof (recu : SIG%d)\n", i);
  else
    printf("Closing spoof (received : SIG%d)\n", i);
  lcrzo_spoof_close(&lcrzoex_000246_spoof);
  raise(SIGTERM);
}

/*-------------------------------------------------------------*/
void lcrzoex_000246_sig_received2(int i)
{ int status, language;

  lcrzo_global_get_language(&language);
  if ( language==LCRZO_GLOBAL_LANG_FR )
    printf("Fermeture du sniff (recu : SIG%d)\n", i);
  else
    printf("Closing sniff (received : SIG%d)\n", i);
  kill(lcrzoex_000246_pid, SIGINT);
  waitpid(lcrzoex_000246_pid, &status, 0);
  lcrzo_sniff_close(&lcrzoex_000246_sniff);
  raise(SIGTERM);
}

/*-------------------------------------------------------------*/
int lcrzoex_000246_sendecho(lcrzo_device device,
			    lcrzo_etha clientetha,
			    lcrzo_etha routeretha,
			    lcrzo_ipl clientipl, 
			    lcrzo_ipl serveripl, 
			    lcrzo_uint16 startingid,
			    lcrzo_uint8 maxttl)
{ lcrzo_hdrleth hdrleth;
  lcrzo_hdrlip hdrlip;
  lcrzo_data data;
  lcrzo_int32 datasize; 
  lcrzo_uint8 i;

  /* initialize Ethernet header */
  lcrzo_epr(lcrzo_hdrleth_initdefault(&hdrleth));
  memcpy(hdrleth.src, clientetha, LCRZO_ETHA_MAXBYTES);
  memcpy(hdrleth.dst, routeretha, LCRZO_ETHA_MAXBYTES);

  /* initialize IP header */
  lcrzo_epr(lcrzo_hdrlip_initdefault(&hdrlip));
  hdrlip.saddr=clientipl;
  hdrlip.daddr=serveripl;
  hdrlip.ttl=1;
  hdrlip.id=startingid;

  /* spoof */
  lcrzo_epr(lcrzo_spoof_init(&lcrzoex_000246_spoof));
  /*lcrzo_epr(lcrzo_spoof_set_printbeforesending(&lcrzoex_000246_spoof,
                                               LCRZO_TRUE));*/
  for ( i=1 ; i<=maxttl ; i++ )
  { lcrzo_epr(lcrzo_packet_initm_ethipopticmpecho(hdrleth,
						  hdrlip, NULL, 0,
						  LCRZO_ICMPTYPE_ECHOREQ,
						  startingid, startingid,
						  "ICMP traceroute", 15,
						  &data, &datasize));
    lcrzo_epr(lcrzo_spoof_eth(&lcrzoex_000246_spoof, device, data, datasize));
    lcrzo_epr(lcrzo_data_free(data));
    hdrlip.id++;
    hdrlip.ttl++;
    startingid++;
    lcrzo_epr(lcrzo_time_sleep_sec(1));
  }
  lcrzo_epr(lcrzo_spoof_close(&lcrzoex_000246_spoof));
  
  return(LCRZO_ERR_OK);
}

/*-------------------------------------------------------------*/
int lcrzoex_000246_print(lcrzo_device device,
			 lcrzo_etha clientetha,
			 lcrzo_etha routeretha,
			 lcrzo_ipl clientipl, 
			 lcrzo_ipl serveripl, 
			 lcrzo_uint16 startingid,
			 lcrzo_uint8 maxttl)
{ lcrzo_uint32 mtu;
  lcrzo_data sniffdata;
  lcrzo_int32 sniffdatasize;
  int ret;

  /* init sniff */
  lcrzo_epr(lcrzo_mtu_init_device(device, &mtu));
  lcrzo_epr(lcrzo_sniff_init(device, mtu, "", &lcrzoex_000246_sniff));

  /* main loop */
  ret=lcrzo_sniff_nextm(&lcrzoex_000246_sniff, LCRZO_TRUE, 
			LCRZO_SNIFF_TYPE_NORMAL, &sniffdata, &sniffdatasize);
  while (ret==LCRZO_ERR_OK)
  { /* answer to requests */
    lcrzo_epr(lcrzo_virtual_answer_alive(sniffdata, sniffdatasize, 
                                         device, clientetha, clientipl));
    /* decode an icmp error, and print it */
    lcrzo_epr(lcrzoex_000246_printroutei(clientetha, routeretha,
					 clientipl, serveripl, startingid,
					 maxttl, sniffdata, sniffdatasize));
    /* decode an icmp echo reply, and print it */
    lcrzo_epr(lcrzoex_000246_printrouter(clientetha, routeretha,
					 clientipl, serveripl, startingid,
					 sniffdata, sniffdatasize));
    lcrzo_data_free(sniffdata);
    /* sniff next packet */
    ret=lcrzo_sniff_nextm(&lcrzoex_000246_sniff, LCRZO_TRUE,
			  LCRZO_SNIFF_TYPE_NORMAL, &sniffdata, &sniffdatasize);
  }

  /* close everything */
  raise(SIGINT);
  /*never reatched*/
  return(LCRZO_ERR_IEINTERNALERROR);
}

/*-------------------------------------------------------------*/
int lcrzoex_000246_printroutei(lcrzo_etha clientetha,
			       lcrzo_etha routeretha,
			       lcrzo_ipl clientipl, 
			       lcrzo_ipl serveripl, 
			       lcrzo_uint16 startingid,
			       lcrzo_uint8 maxttl,
			       lcrzo_data sniffdata,
			       lcrzo_int32 sniffdatasize)
{ lcrzo_hdrleth hdrleth;
  lcrzo_hdrlip hdrlip, badhdrip;
  lcrzo_data baddata;
  lcrzo_string string;
  lcrzo_uint16 baddatasize;
  lcrzo_uint8 type, code, badicmptype, badicmpcode, b4, b5, b6, b7;
  int ret, indexidok, indexidko, indexechoid, indexechoseqnum, language;

  /*lcrzo_epr(lcrzo_packet_print_eth(sniffdata, sniffdatasize,
				   LCRZO_PRINTPROFILE_ARRAYDUMP));*/

  /* try to decode icmp */
  ret=lcrzo_packet_decodem_ethipopticmperror(sniffdata, sniffdatasize,
					     &hdrleth, &hdrlip, NULL,NULL,
					     &type, &code, NULL, NULL,
					     &baddata, &baddatasize);
  if ( ret!=LCRZO_ERR_OK )
    return(LCRZO_ERR_OK);

  /* check if the addresses are the one we are looking for */
  if ( ! lcrzo_etha_equal(hdrleth.src, routeretha) ||
       ! lcrzo_etha_equal(hdrleth.dst, clientetha) ||
       ! lcrzo_ipl_equal(hdrlip.daddr, clientipl) )
  { lcrzo_data_free(baddata);
    return(LCRZO_ERR_OK);
  }

  /* icmp should contain icmp */
  ret=lcrzo_ip64bits_decodem_ipicmp(baddata, baddatasize,
				    &badhdrip, &badicmptype, &badicmpcode,
				    NULL, &b4, &b5, &b6, &b7);
  lcrzo_data_free(baddata);
  if ( ret!=LCRZO_ERR_OK )
    return(LCRZO_ERR_OK);

  /* check various values in baddata */
  if ( badicmptype!=LCRZO_ICMPTYPE_ECHOREQ ||
       badicmpcode!=0 ||
       badhdrip.saddr!=clientipl ||
       badhdrip.daddr!=serveripl )
    return(LCRZO_ERR_OK);

  /* compute indexid from the id of the received packet */
  indexidok=badhdrip.id-startingid+1;
  /* this one is because there is a bug in the IP stack of BSD :
     id is reversed :( */
  indexidko= ( (badhdrip.id&0xFF)<<8 ) | ( (badhdrip.id&0xFF00)>>8 );
  indexidko=indexidko-startingid+1;
  /* compute indexechoid from the id of the received packet */
  indexechoid=((b4<<8)|b5)-startingid+1;
  /* compute indexechoseqnum from the seqnum of the received packet */
  indexechoseqnum=((b6<<8)|b7)-startingid+1;

  /* verify that indexes are equal */
  if ( (indexidok!=indexechoid&&indexidko!=indexechoid) ||
       indexechoid!=indexechoseqnum )
    return(LCRZO_ERR_OK);

  /* print the result */
  printf("% 3d : ", indexechoid);
  lcrzo_epr(lcrzo_ipl_print("", hdrlip.saddr, ""));
      
  /* print the type of error */
  if ( type==LCRZO_ICMPTYPE_TIMEEXCEED &&
       code == LCRZO_ICMPCODE_TIMEEXCEED_TTL )
  { /* do not display the error type and code */
    printf("\n"); 
  }
  else 
  { /* display the error type and code */
    lcrzo_epr(lcrzo_string_initm_icmptypecode(type, code, &string));
    printf(" : %s\n", string);
    lcrzo_string_free(string);
  }

  /* if maxttl was too short */
  if ( indexechoid == maxttl )
  { lcrzo_epr(lcrzo_global_get_language(&language));
    if ( language==LCRZO_GLOBAL_LANG_FR )
      printf("maxttl(%d) est trop petit pour atteindre la destination\n", 
	     maxttl);
    else
      printf("maxttl(%d) was too short to reach the destination\n", 
	     maxttl);
    raise(SIGINT);
  }

  return(LCRZO_ERR_OK);
}

/*-------------------------------------------------------------*/
int lcrzoex_000246_printrouter(lcrzo_etha clientetha,
			       lcrzo_etha routeretha,
			       lcrzo_ipl clientipl, 
			       lcrzo_ipl serveripl, 
			       lcrzo_uint16 startingid,
			       lcrzo_data sniffdata,
			       lcrzo_int32 sniffdatasize)
{ static lcrzo_bool destreatched=LCRZO_FALSE;
  lcrzo_hdrleth hdrleth;
  lcrzo_hdrlip hdrlip;
  lcrzo_uint8 type;
  lcrzo_uint16 id, seqnum;
  int ret, theindex1, theindex2, language;

  /*lcrzo_epr(lcrzo_packet_print_eth(sniffdata, sniffdatasize,
    LCRZO_PRINTPROFILE_ARRAYDUMP));*/

  /* try to decode it */
  ret=lcrzo_packet_decodem_ethipopticmpecho(sniffdata, sniffdatasize,
					    &hdrleth, &hdrlip, NULL, NULL,
					    &type, &id, &seqnum, NULL, NULL);
  if ( ret!=LCRZO_ERR_OK )
    return(LCRZO_ERR_OK);

  /* check if the addresses are the one we are looking for */
  if ( ! lcrzo_etha_equal(hdrleth.src, routeretha) || 
       ! lcrzo_etha_equal(hdrleth.dst, clientetha) || 
       ! lcrzo_ipl_equal(hdrlip.saddr, serveripl) ||
       ! lcrzo_ipl_equal(hdrlip.daddr, clientipl) )
    return(LCRZO_ERR_OK);

  /* check type */
  if ( type!=LCRZO_ICMPTYPE_ECHOREP )
    return(LCRZO_ERR_OK);

  /* compute the theindex from the received packet */
  theindex1=id-startingid+1;
  theindex2=seqnum-startingid+1;
  if ( theindex1!=theindex2 )
    return(LCRZO_ERR_OK);

  /* destination is reatched */
  lcrzo_epr(lcrzo_global_get_language(&language));
  if ( !destreatched )
  { destreatched=LCRZO_TRUE;
    printf("% 3d : ", theindex1);
    lcrzo_epr(lcrzo_ipl_print("", hdrlip.saddr, "\n"));
    if ( language==LCRZO_GLOBAL_LANG_FR )
    { printf("Destination atteinte.\n");
      printf("On se donne 3 secondes pour attendre les eventuelles ");
      printf("reponses retardataires.\n");
    }
    else
    { printf("Destination reatched.\n");
      printf("We continue during 3 seconds to wait for late answers.\n");
    }
    signal(SIGALRM, lcrzoex_000246_sig_received2);
    alarm(3);
  }

  return(LCRZO_ERR_OK);
}
