
/* include the library headers */
#include <lcrzo.h>
#include "lcrzoex.h"
#include "shared/shared_misc.h"

/*-------------------------------------------------------------*/
/* set information about this example */
lcrzoex_infos lcrzoex_000248=
{ /*reference*/     "lcrzoex", 248,
  /*version*/       1, 0,
  /*title*/         "icmp ping (spoofed)",
  /*french title*/  "ping icmp (spoofe)",
  /*description*/   "",
  /*french desc.*/  "",
  /*usage*/         "device srceth routereth srcip dstip",
  /*french usage*/  "device ethsrc ethrouteur ipsrc ipdst",
  /*usage example*/ "eth0 aa:bb:cc:dd:ee:ff 11:22:33:44:55:66 1.2.3.4 5.6.7.78",
  /*fr. usage ex.*/ "eth0 aa:bb:cc:dd:ee:ff 11:22:33:44:55:66 1.2.3.4 5.6.7.78",
  /*author*/        "Laurent"
};

/*-------------------------------------------------------------*/
/* function declarations */
int lcrzoex_000248_main(int argc, char *argv[]);
void lcrzoex_000248_sig_received1(int i);
void lcrzoex_000248_sig_received2(int i);
int lcrzoex_000248_ping(lcrzo_device device,
			lcrzo_etha clientetha,
			lcrzo_etha routeretha,
			lcrzo_ipl clientipl, 
			lcrzo_ipl serveripl);
int lcrzoex_000248_sendecho(lcrzo_device device,
			    lcrzo_etha clientetha,
			    lcrzo_etha routeretha,
			    lcrzo_ipl clientipl, 
			    lcrzo_ipl serveripl, 
			    lcrzo_uint16 startingid);
int lcrzoex_000248_print(lcrzo_device device,
			 lcrzo_etha clientetha,
			 lcrzo_ipl clientipl, 
			 lcrzo_ipl serveripl, 
			 lcrzo_uint16 startingid);
int lcrzoex_000248_printerror(lcrzo_ipl clientipl, 
			      lcrzo_ipl serveripl, 
			      lcrzo_uint16 startingid,
			      lcrzo_data sniffdata,
			      lcrzo_int32 sniffdatasize);
int lcrzoex_000248_printreply(lcrzo_ipl clientipl, 
			      lcrzo_uint16 startingid,
			      lcrzo_data sniffdata,
			      lcrzo_int32 sniffdatasize);

/*-------------------------------------------------------------*/
/* global variables */
int lcrzoex_000248_pid;
lcrzo_spoof lcrzoex_000248_spoof;
lcrzo_sniff lcrzoex_000248_sniff;

/*-------------------------------------------------------------*/
/* main function : you can rename it to "main", and then this file
   can be compiled alone. */
int lcrzoex_000248_main(int argc, char *argv[])
{ lcrzo_device device;
  lcrzo_etha clientetha, routeretha;
  lcrzo_ipl clientipl, serveripl;

  if ( argc<6 )
  { /* print usage */
    lcrzo_epr(lcrzoex_infos_print(lcrzoex_000248, argv[0]));
    return(LCRZO_ERR_BULCRZOEXBADUSAGE);
  }
  lcrzo_epr(lcrzo_device_init(argv[1], device));
  lcrzo_epr(lcrzo_etha_init_eths(argv[2], clientetha));
  lcrzo_epr(lcrzo_etha_init_eths(argv[3], routeretha));
  lcrzo_epr(lcrzo_ipl_init_hs(argv[4], &clientipl));
  lcrzo_epr(lcrzo_ipl_init_hs(argv[5], &serveripl));

  /* call the main function */
  lcrzo_epr(lcrzoex_000248_ping(device, clientetha, routeretha,
				clientipl, serveripl));

  return (LCRZO_ERR_OK);
}

/*-------------------------------------------------------------*/
int lcrzoex_000248_ping(lcrzo_device device,
			lcrzo_etha clientetha,
			lcrzo_etha routeretha,
			lcrzo_ipl clientipl, 
			lcrzo_ipl serveripl)
{ lcrzo_uint16 startingid;

  /* determine starting values */
  lcrzo_epr(lcrzo_uint16_rand(5000, 30000, &startingid));

  lcrzoex_000248_pid=fork();
  if (lcrzoex_000248_pid<0) return(LCRZO_ERR_FUFORK);
  if (lcrzoex_000248_pid==0)
  { /* this process sends an echo request */
    /* in case we receive a signal */
    signal(SIGINT, lcrzoex_000248_sig_received1);
    lcrzo_epr(lcrzoex_000248_sendecho(device, clientetha, routeretha,
				      clientipl, serveripl, startingid));
  }
  else
  { /* this process sniff and prints ip addresses */
    /* in case we receive a signal */
    signal(SIGINT, lcrzoex_000248_sig_received2);
    lcrzo_epr(lcrzoex_000248_print(device, clientetha, clientipl, 
				   serveripl, startingid));
  }
  return (LCRZO_ERR_OK);
}

/*-------------------------------------------------------------*/
void lcrzoex_000248_sig_received1(int i)
{ int language;

  lcrzo_global_get_language(&language);
  if ( language==LCRZO_GLOBAL_LANG_FR )
    printf("Fermeture du spoof (recu : SIG%d)\n", i);
  else
    printf("Closing spoof (received : SIG%d)\n", i);
  lcrzo_spoof_close(&lcrzoex_000248_spoof);
  raise(SIGTERM);
}

/*-------------------------------------------------------------*/
void lcrzoex_000248_sig_received2(int i)
{ int status, language;

  lcrzo_global_get_language(&language);
  if ( language==LCRZO_GLOBAL_LANG_FR )
    printf("Fermeture du sniff (recu : SIG%d)\n", i);
  else
    printf("Closing sniff (received : SIG%d)\n", i);
  kill(lcrzoex_000248_pid, SIGINT);
  waitpid(lcrzoex_000248_pid, &status, 0);
  lcrzo_sniff_close(&lcrzoex_000248_sniff);
  raise(SIGTERM);
}

/*-------------------------------------------------------------*/
int lcrzoex_000248_sendecho(lcrzo_device device,
			    lcrzo_etha clientetha,
			    lcrzo_etha routeretha,
			    lcrzo_ipl clientipl, 
			    lcrzo_ipl serveripl, 
			    lcrzo_uint16 startingid)
{ lcrzo_hdrleth hdrleth;
  lcrzo_hdrlip hdrlip;
  lcrzo_data data;
  lcrzo_int32 datasize; 

  /* initialize Ethernet header */
  lcrzo_epr(lcrzo_hdrleth_initdefault(&hdrleth));
  memcpy(hdrleth.src, clientetha, LCRZO_ETHA_MAXBYTES);
  memcpy(hdrleth.dst, routeretha, LCRZO_ETHA_MAXBYTES);

  /* initialize IP header */
  lcrzo_epr(lcrzo_hdrlip_initdefault(&hdrlip));
  hdrlip.saddr=clientipl;
  hdrlip.daddr=serveripl;
  hdrlip.id=startingid;

  /* spoof */
  lcrzo_epr(lcrzo_spoof_init(&lcrzoex_000248_spoof));
  /*lcrzo_epr(lcrzo_spoof_set_printbeforesending(&lcrzoex_000248_spoof,
                                               LCRZO_TRUE));*/
  while(LCRZO_TRUE)
  { lcrzo_epr(lcrzo_packet_initm_ethipopticmpecho(hdrleth,
						  hdrlip, NULL, 0,
						  LCRZO_ICMPTYPE_ECHOREQ,
						  startingid, startingid,
						  "ICMP ping", 9,
						  &data, &datasize));
    lcrzo_epr(lcrzo_spoof_eth(&lcrzoex_000248_spoof, device, data, datasize));
    lcrzo_epr(lcrzo_data_free(data));
    hdrlip.id++;
    startingid++;
    lcrzo_epr(lcrzo_time_sleep_sec(1));
  }
  lcrzo_epr(lcrzo_spoof_close(&lcrzoex_000248_spoof));
  
  return(LCRZO_ERR_OK);
}

/*-------------------------------------------------------------*/
int lcrzoex_000248_print(lcrzo_device device,
			 lcrzo_etha clientetha,
			 lcrzo_ipl clientipl, 
			 lcrzo_ipl serveripl, 
			 lcrzo_uint16 startingid)
{ lcrzo_uint32 mtu;
  lcrzo_data sniffdata;
  lcrzo_int32 sniffdatasize;
  int ret;

  /* init sniff */
  lcrzo_epr(lcrzo_mtu_init_device(device, &mtu));
  lcrzo_epr(lcrzo_sniff_init(device, mtu, "", &lcrzoex_000248_sniff));

  /* main loop */
  ret=lcrzo_sniff_nextm(&lcrzoex_000248_sniff, LCRZO_TRUE, 
			LCRZO_SNIFF_TYPE_NORMAL, &sniffdata, &sniffdatasize);
  while (ret==LCRZO_ERR_OK)
  { /* answer to requests */
    lcrzo_epr(lcrzo_virtual_answer_alive(sniffdata, sniffdatasize, 
                                         device, clientetha, clientipl));
    /* decode an icmp error, and print it */
    lcrzo_epr(lcrzoex_000248_printerror(clientipl, serveripl, startingid,
					sniffdata, sniffdatasize));
    /* decode an icmp echo reply, and print it */
    lcrzo_epr(lcrzoex_000248_printreply(clientipl, startingid,
					sniffdata, sniffdatasize));
    lcrzo_data_free(sniffdata);
    /* sniff next packet */
    ret=lcrzo_sniff_nextm(&lcrzoex_000248_sniff, LCRZO_TRUE,
			  LCRZO_SNIFF_TYPE_NORMAL, &sniffdata, &sniffdatasize);
  }

  /* close everything */
  raise(SIGINT);
  /*never reatched*/
  return(LCRZO_ERR_IEINTERNALERROR);
}

/*-------------------------------------------------------------*/
int lcrzoex_000248_printerror(lcrzo_ipl clientipl, 
			      lcrzo_ipl serveripl, 
			      lcrzo_uint16 startingid,
			      lcrzo_data sniffdata,
			      lcrzo_int32 sniffdatasize)
{ lcrzo_hdrleth hdrleth;
  lcrzo_hdrlip hdrlip, badhdrip;
  lcrzo_data baddata;
  lcrzo_string string;
  lcrzo_uint16 baddatasize;
  lcrzo_uint8 type, code, badicmptype, badicmpcode, b4, b5, b6, b7;
  int ret, indexidok, indexidko, indexechoid, indexechoseqnum;

  /*lcrzo_epr(lcrzo_packet_print_eth(sniffdata, sniffdatasize,
				   LCRZO_PRINTPROFILE_ARRAYDUMP));*/

  /* try to decode icmp */
  ret=lcrzo_packet_decodem_ethipopticmperror(sniffdata, sniffdatasize,
					     &hdrleth, &hdrlip, NULL,NULL,
					     &type, &code, NULL, NULL,
					     &baddata, &baddatasize);
  if ( ret!=LCRZO_ERR_OK )
    return(LCRZO_ERR_OK);

  /* check if the addresses are the one we are looking for */
  if ( ! lcrzo_ipl_equal(hdrlip.daddr, clientipl) )
  { lcrzo_data_free(baddata);
    return(LCRZO_ERR_OK);
  }

  /* icmp should contain icmp */
  ret=lcrzo_ip64bits_decodem_ipicmp(baddata, baddatasize,
				    &badhdrip, &badicmptype, &badicmpcode,
				    NULL, &b4, &b5, &b6, &b7);
  lcrzo_data_free(baddata);
  if ( ret!=LCRZO_ERR_OK )
    return(LCRZO_ERR_OK);

  /* check various values in baddata */
  if ( badicmptype!=LCRZO_ICMPTYPE_ECHOREQ ||
       badicmpcode!=0 ||
       badhdrip.saddr!=clientipl ||
       badhdrip.daddr!=serveripl )
    return(LCRZO_ERR_OK);

  /* compute indexid from the id of the received packet */
  indexidok=badhdrip.id-startingid+1;
  /* this one is because there is a bug in the IP stack of BSD :
     id is reversed :( */
  indexidko= ( (badhdrip.id&0xFF)<<8 ) | ( (badhdrip.id&0xFF00)>>8 );
  indexidko=indexidko-startingid+1;
  /* compute indexechoid from the id of the received packet */
  indexechoid=((b4<<8)|b5)-startingid+1;
  /* compute indexechoseqnum from the seqnum of the received packet */
  indexechoseqnum=((b6<<8)|b7)-startingid+1;

  /* verify that indexes are equal */
  if ( (indexidok!=indexechoid&&indexidko!=indexechoid) ||
       indexechoid!=indexechoseqnum )
    return(LCRZO_ERR_OK);

  /* display the error type and code */
  printf("% 3d : ", indexechoid);
  lcrzo_epr(lcrzo_ipl_print("", hdrlip.saddr, ""));
  lcrzo_epr(lcrzo_string_initm_icmptypecode(type, code, &string));
  printf(" : %s\n", string);
  lcrzo_string_free(string);

  return(LCRZO_ERR_OK);
}

/*-------------------------------------------------------------*/
int lcrzoex_000248_printreply(lcrzo_ipl clientipl, 
			      lcrzo_uint16 startingid,
			      lcrzo_data sniffdata,
			      lcrzo_int32 sniffdatasize)
{ lcrzo_hdrleth hdrleth;
  lcrzo_hdrlip hdrlip;
  lcrzo_uint8 type;
  lcrzo_uint16 id, seqnum;
  int ret, theindex1, theindex2;

  /*lcrzo_epr(lcrzo_packet_print_eth(sniffdata, sniffdatasize,
				   LCRZO_PRINTPROFILE_ARRAYDUMP));*/

  /* try to decode it */
  ret=lcrzo_packet_decodem_ethipopticmpecho(sniffdata, sniffdatasize,
					    &hdrleth, &hdrlip, NULL, NULL,
					    &type, &id, &seqnum, NULL, NULL);
  if ( ret!=LCRZO_ERR_OK )
    return(LCRZO_ERR_OK);

  /* check if the addresses are the one we are looking for */
  if ( ! lcrzo_ipl_equal(hdrlip.daddr, clientipl) )
    return(LCRZO_ERR_OK);

  /* check type */
  if ( type!=LCRZO_ICMPTYPE_ECHOREP )
    return(LCRZO_ERR_OK);

  /* compute the theindex from the received packet */
  theindex1=id-startingid+1;
  theindex2=seqnum-startingid+1;
  if ( theindex1!=theindex2 )
    return(LCRZO_ERR_OK);

  /* destination is reatched */
  printf("% 3d : ", theindex1);
  lcrzo_epr(lcrzo_etha_print("", hdrleth.src, "\t"));
  lcrzo_epr(lcrzo_ipl_print("", hdrlip.saddr, "\n"));

  return(LCRZO_ERR_OK);
}
