
/* include the library headers */
#include <lcrzo.h>
#include "lcrzoex.h"
#include "shared/shared_misc.h"

/*-------------------------------------------------------------*/
/* set information about this example */
lcrzoex_infos lcrzoex_000250=
{ /*reference*/     "lcrzoex", 250,
  /*version*/       1, 0,
  /*title*/         "tcp ping (spoofed)",
  /*french title*/  "ping tcp (spoofe)",
  /*description*/   "",
  /*french desc.*/  "",
  /*usage*/         "device srceth routereth srcip dstip dstport",
  /*french usage*/  "device ethsrc ethrouteur ipsrc ipdst portdst",
  /*usage example*/ "eth0 aa:bb:cc:dd:ee:ff 11:22:33:44:55:66 1.2.3.4 5.6.7.78 80",
  /*fr. usage ex.*/ "eth0 aa:bb:cc:dd:ee:ff 11:22:33:44:55:66 1.2.3.4 5.6.7.78 80",
  /*author*/        "Laurent"
};

/*-------------------------------------------------------------*/
/* function declarations */
int lcrzoex_000250_main(int argc, char *argv[]);
void lcrzoex_000250_sig_received1(int i);
void lcrzoex_000250_sig_received2(int i);
int lcrzoex_000250_tcpping(lcrzo_device device,
			   lcrzo_etha clientetha,
			   lcrzo_etha routeretha,
			   lcrzo_ipl clientipl, 
			   lcrzo_ipl serveripl, 
			   lcrzo_uint16 serverport);
int lcrzoex_000250_sendsyn(lcrzo_device device,
			   lcrzo_etha clientetha,
			   lcrzo_etha routeretha,
			   lcrzo_ipl clientipl, 
			   lcrzo_ipl serveripl, 
			   lcrzo_uint16 serverport,
			   lcrzo_uint16 startingclientport,
			   lcrzo_uint32 startingclientseqnum);
int lcrzoex_000250_print(lcrzo_device device,
			 lcrzo_etha clientetha,
			 lcrzo_etha routeretha,
			 lcrzo_ipl clientipl, 
			 lcrzo_ipl serveripl, 
			 lcrzo_uint16 serverport,
			 lcrzo_uint16 startingclientport,
			 lcrzo_uint32 startingclientseqnum);
int lcrzoex_000250_printroutei(lcrzo_etha routeretha,
			       lcrzo_ipl clientipl, 
			       lcrzo_ipl serveripl, 
			       lcrzo_uint16 serverport,
			       lcrzo_uint16 startingclientport,
			       lcrzo_uint32 startingclientseqnum,
			       lcrzo_data sniffdata,
			       lcrzo_int32 sniffdatasize);
int lcrzoex_000250_printroutet(lcrzo_etha routeretha,
			       lcrzo_ipl clientipl, 
			       lcrzo_ipl serveripl, 
			       lcrzo_uint16 serverport,
			       lcrzo_uint16 startingclientport,
			       lcrzo_uint32 startingclientseqnum,
			       lcrzo_data sniffdata,
			       lcrzo_int32 sniffdatasize);

/*-------------------------------------------------------------*/
/* global variables */
int lcrzoex_000250_pid;
lcrzo_spoof lcrzoex_000250_spoof;
lcrzo_sniff lcrzoex_000250_sniff;

/*-------------------------------------------------------------*/
/* main function : you can rename it to "main", and then this file
   can be compiled alone. */
int lcrzoex_000250_main(int argc, char *argv[])
{ lcrzo_device device;
  lcrzo_etha clientetha, routeretha;
  lcrzo_ipl clientipl, serveripl;
  lcrzo_uint16 serverport;

  if ( argc<7 )
  { /* print usage */
    lcrzo_epr(lcrzoex_infos_print(lcrzoex_000250, argv[0]));
    return(LCRZO_ERR_BULCRZOEXBADUSAGE);
  }
  lcrzo_epr(lcrzo_device_init(argv[1], device));
  lcrzo_epr(lcrzo_etha_init_eths(argv[2], clientetha));
  lcrzo_epr(lcrzo_etha_init_eths(argv[3], routeretha));
  lcrzo_epr(lcrzo_ipl_init_hs(argv[4], &clientipl));
  lcrzo_epr(lcrzo_ipl_init_hs(argv[5], &serveripl));
  serverport=(lcrzo_uint16)atoi(argv[6]);

  /* call the main function */
  lcrzo_epr(lcrzoex_000250_tcpping(device, clientetha, routeretha,
				   clientipl, serveripl, serverport));

  return (LCRZO_ERR_OK);
}

/*-------------------------------------------------------------*/
int lcrzoex_000250_tcpping(lcrzo_device device,
			   lcrzo_etha clientetha,
			   lcrzo_etha routeretha,
			   lcrzo_ipl clientipl, 
			   lcrzo_ipl serveripl, 
			   lcrzo_uint16 serverport)
{ lcrzo_uint16 startingclientport;
  lcrzo_uint32 startingclientseqnum;

  /* determine starting values */
  lcrzo_epr(lcrzo_uint16_rand(5000, 30000, &startingclientport));
  lcrzo_epr(lcrzo_uint32_rand(0, 0xEFFFFFFFU, &startingclientseqnum));

  lcrzoex_000250_pid=fork();
  if (lcrzoex_000250_pid<0) return(LCRZO_ERR_FUFORK);
  if (lcrzoex_000250_pid==0)
  { /* this process sends a SYN */
    /* in case we receive a signal */
    signal(SIGINT, lcrzoex_000250_sig_received1);
    lcrzo_epr(lcrzoex_000250_sendsyn(device, clientetha, routeretha,
				     clientipl, serveripl, serverport,
				     startingclientport, 
				     startingclientseqnum));
  }
  else
  { /* this process sniff and prints sequence numbers */
    /* in case we receive a signal */
    signal(SIGINT, lcrzoex_000250_sig_received2);
    lcrzo_epr(lcrzoex_000250_print(device, clientetha, routeretha,
				   clientipl, serveripl, serverport,
				   startingclientport, startingclientseqnum));
  }
  return (LCRZO_ERR_OK);
}

/*-------------------------------------------------------------*/
void lcrzoex_000250_sig_received1(int i)
{ int language;

  lcrzo_global_get_language(&language);
  if ( language==LCRZO_GLOBAL_LANG_FR )
    printf("Fermeture du spoof (recu : SIG%d)\n", i);
  else
    printf("Closing spoof (received : SIG%d)\n", i);
  lcrzo_spoof_close(&lcrzoex_000250_spoof);
  raise(SIGTERM);
}

/*-------------------------------------------------------------*/
void lcrzoex_000250_sig_received2(int i)
{ int status, language;

  lcrzo_global_get_language(&language);
  if ( language==LCRZO_GLOBAL_LANG_FR )
    printf("Fermeture du sniff (recu : SIG%d)\n", i);
  else
    printf("Closing sniff (received : SIG%d)\n", i);
  kill(lcrzoex_000250_pid, SIGINT);
  waitpid(lcrzoex_000250_pid, &status, 0);
  lcrzo_sniff_close(&lcrzoex_000250_sniff);
  raise(SIGTERM);
}

/*-------------------------------------------------------------*/
int lcrzoex_000250_sendsyn(lcrzo_device device,
			   lcrzo_etha clientetha,
			   lcrzo_etha routeretha,
			   lcrzo_ipl clientipl, 
			   lcrzo_ipl serveripl, 
			   lcrzo_uint16 serverport,
			   lcrzo_uint16 startingclientport,
			   lcrzo_uint32 startingclientseqnum)
{ lcrzo_hdrleth hdrleth;
  lcrzo_hdrlip hdrlip;
  lcrzo_hdrltcp hdrltcp;

  /* initialize Ethernet header */
  lcrzo_epr(lcrzo_hdrleth_initdefault(&hdrleth));
  memcpy(hdrleth.src, clientetha, LCRZO_ETHA_MAXBYTES);
  memcpy(hdrleth.dst, routeretha, LCRZO_ETHA_MAXBYTES);

  /* initialize IP header */
  lcrzo_epr(lcrzo_hdrlip_initdefault(&hdrlip));
  hdrlip.saddr=clientipl;
  hdrlip.daddr=serveripl;
  hdrlip.id=startingclientport;

  /* initialize TCP header */
  lcrzo_epr(lcrzo_hdrltcp_initdefault(&hdrltcp));
  hdrltcp.dport=serverport;
  hdrltcp.syn=1;
  hdrltcp.sport=startingclientport;
  hdrltcp.seqnum=startingclientseqnum;

  /* spoof */
  lcrzo_epr(lcrzo_spoof_init(&lcrzoex_000250_spoof));
  /*lcrzo_epr(lcrzo_spoof_set_printbeforesending(&lcrzoex_000250_spoof,
                                               LCRZO_TRUE));*/
  while(LCRZO_TRUE)
  { lcrzo_epr(lcrzo_spoof_ethipopttcpoptdata(&lcrzoex_000250_spoof,
					     device, hdrleth,
					     hdrlip, NULL, 0,
					     hdrltcp, NULL, 0,
					     NULL, 0));
    hdrlip.id++;
    hdrltcp.sport++;
    hdrltcp.seqnum+=10;
    lcrzo_epr(lcrzo_time_sleep_sec(1));
  }
  lcrzo_epr(lcrzo_spoof_close(&lcrzoex_000250_spoof));
  
  return(LCRZO_ERR_OK);
}

/*-------------------------------------------------------------*/
int lcrzoex_000250_print(lcrzo_device device,
			 lcrzo_etha clientetha,
			 lcrzo_etha routeretha,
			 lcrzo_ipl clientipl, 
			 lcrzo_ipl serveripl, 
			 lcrzo_uint16 serverport,
			 lcrzo_uint16 startingclientport,
			 lcrzo_uint32 startingclientseqnum)
{ lcrzo_uint32 mtu;
  lcrzo_data sniffdata;
  lcrzo_int32 sniffdatasize;
  int ret;

  /* init sniff */
  lcrzo_epr(lcrzo_mtu_init_device(device, &mtu));
  lcrzo_epr(lcrzo_sniff_init(device, mtu, "", &lcrzoex_000250_sniff));

  /* main loop */
  ret=lcrzo_sniff_nextm(&lcrzoex_000250_sniff, LCRZO_TRUE, 
			LCRZO_SNIFF_TYPE_NORMAL, &sniffdata, &sniffdatasize);
  while (ret==LCRZO_ERR_OK)
  { /* answer to requests */
    lcrzo_epr(lcrzo_virtual_answer_alive(sniffdata, sniffdatasize, 
                                         device, clientetha, clientipl));
    /* decode an icmp error, and print it */
    lcrzo_epr(lcrzoex_000250_printroutei(routeretha,
					 clientipl, serveripl, serverport,
					 startingclientport, 
					 startingclientseqnum,
					 sniffdata, sniffdatasize));
    /* decode a tcp synack or reset, and print it */
    lcrzo_epr(lcrzoex_000250_printroutet(routeretha,
					 clientipl, serveripl, serverport,
					 startingclientport, 
					 startingclientseqnum,
					 sniffdata, sniffdatasize));
    lcrzo_data_free(sniffdata);
    /* sniff next packet */
    ret=lcrzo_sniff_nextm(&lcrzoex_000250_sniff, LCRZO_TRUE,
			  LCRZO_SNIFF_TYPE_NORMAL, &sniffdata, &sniffdatasize);
  }

  /* close everything */
  raise(SIGINT);
  /*never reatched*/
  return(LCRZO_ERR_IEINTERNALERROR);
}

/*-------------------------------------------------------------*/
int lcrzoex_000250_printroutei(lcrzo_etha routeretha,
			       lcrzo_ipl clientipl, 
			       lcrzo_ipl serveripl, 
			       lcrzo_uint16 serverport,
			       lcrzo_uint16 startingclientport,
			       lcrzo_uint32 startingclientseqnum,
			       lcrzo_data sniffdata,
			       lcrzo_int32 sniffdatasize)
{ lcrzo_hdrleth hdrleth;
  lcrzo_hdrlip hdrlip, badhdrip;
  lcrzo_data baddata;
  lcrzo_string string;
  lcrzo_uint16 baddatasize, badtcpsport, badtcpdport;
  lcrzo_uint8 type, code;
  lcrzo_uint32 badtcpseqnum;
  int ret, indexidok, indexidko, indexsport, indexseqnum;

  /*lcrzo_epr(lcrzo_packet_print_eth(sniffdata, sniffdatasize,
				   LCRZO_PRINTPROFILE_ARRAYDUMP));*/

  /* try to decode icmp */
  ret=lcrzo_packet_decodem_ethipopticmperror(sniffdata, sniffdatasize,
					     &hdrleth, &hdrlip, NULL,NULL,
					     &type, &code, NULL, NULL,
					     &baddata, &baddatasize);
  if ( ret!=LCRZO_ERR_OK )
    return(LCRZO_ERR_OK);

  /* check if the addresses are the one we are looking for */
  if ( ! lcrzo_etha_equal(hdrleth.src, routeretha) ||
       ! lcrzo_ipl_equal(hdrlip.daddr, clientipl) )
  { lcrzo_data_free(baddata);
    return(LCRZO_ERR_OK);
  }

  /* icmp should contain tcp */
  ret=lcrzo_ip64bits_decodem_iptcp(baddata, baddatasize,
				   &badhdrip, &badtcpsport, &badtcpdport,
				   &badtcpseqnum);
  lcrzo_data_free(baddata);
  if ( ret!=LCRZO_ERR_OK )
    return(LCRZO_ERR_OK);

  /* check various values in baddata */
  if ( badtcpdport!=serverport ||
       badhdrip.saddr!=clientipl ||
       badhdrip.daddr!=serveripl )
    return(LCRZO_ERR_OK);

  /* compute indexid from the id of the received packet */
  indexidok=badhdrip.id-startingclientport+1;
  /* this one is because there is a bug in the IP stack of BSD :
     id is reversed :( */
  indexidko= ( (badhdrip.id&0xFF)<<8 ) | ( (badhdrip.id&0xFF00)>>8 );
  indexidko=indexidko-startingclientport+1;
  /* compute indexsport from the sport of the received packet */
  indexsport=badtcpsport-startingclientport+1;
  /* compute indexseqnum from the seqnum of the received packet */
  indexseqnum=1+(badtcpseqnum-startingclientseqnum)/10;

  /* verify that indexes are equal */
  if ( (indexidok!=indexsport&&indexidko!=indexsport) ||
       indexsport!=indexseqnum )
    return(LCRZO_ERR_OK);

  /* display the error type and code */
  printf("% 3d : ", indexsport);
  lcrzo_epr(lcrzo_ipl_print("", hdrlip.saddr, ""));
  lcrzo_epr(lcrzo_string_initm_icmptypecode(type, code, &string));
  printf(" : %s\n", string);
  lcrzo_string_free(string);

  return(LCRZO_ERR_OK);
}

/*-------------------------------------------------------------*/
int lcrzoex_000250_printroutet(lcrzo_etha routeretha,
			       lcrzo_ipl clientipl, 
			       lcrzo_ipl serveripl, 
			       lcrzo_uint16 serverport,
			       lcrzo_uint16 startingclientport,
			       lcrzo_uint32 startingclientseqnum,
			       lcrzo_data sniffdata,
			       lcrzo_int32 sniffdatasize)
{ lcrzo_hdrleth hdrleth;
  lcrzo_hdrlip hdrlip;
  lcrzo_hdrltcp hdrltcp;
  int ret, theindex1, theindex2, language;

  /*lcrzo_epr(lcrzo_packet_print_eth(sniffdata, sniffdatasize,
    LCRZO_PRINTPROFILE_ARRAYDUMP));*/

  /* try to decode it */
  ret=lcrzo_packet_decodem_ethipopttcpoptdata(sniffdata, sniffdatasize,
					      &hdrleth, &hdrlip, NULL, NULL,
					      &hdrltcp, NULL, NULL,
					      NULL, NULL);
  if ( ret!=LCRZO_ERR_OK )
    return(LCRZO_ERR_OK);

  /* check if the addresses are the one we are looking for */
  if ( ! lcrzo_etha_equal(hdrleth.src, routeretha) || 
       ! lcrzo_ipl_equal(hdrlip.saddr, serveripl) ||
       ! lcrzo_ipl_equal(hdrlip.daddr, clientipl) )
    return(LCRZO_ERR_OK);

  /* check server port */
  if ( hdrltcp.sport!=serverport )
    return(LCRZO_ERR_OK);

  /* compute the theindex from the received packet */
  theindex1=hdrltcp.dport-startingclientport+1;
  theindex2=1+(hdrltcp.acknum-1-startingclientseqnum)/10;
  if ( theindex1!=theindex2 )
    return(LCRZO_ERR_OK);

  /* if we received synack or reset, destination is reatched */
  if ( ( hdrltcp.syn && hdrltcp.ack ) || hdrltcp.rst )
  { printf("% 3d : ", theindex1);
    lcrzo_epr(lcrzo_ipl_print("", hdrlip.saddr, ""));
    if (hdrltcp.rst)
    { lcrzo_epr(lcrzo_global_get_language(&language));
      if ( language==LCRZO_GLOBAL_LANG_FR )
        printf(" (le port TCP %d est ferme)\n", serverport);
      else
        printf(" (the TCP port %d is closed)\n", serverport);
    }
    else
    { printf("\n");
    }
  }

  return(LCRZO_ERR_OK);
}
