
/* include the library headers */
#include <lcrzo.h>
#include "lcrzoex.h"
#include "shared/shared_misc.h"

/*-------------------------------------------------------------*/
/* set information about this example */
lcrzoex_infos lcrzoex_000252=
{ /*reference*/     "lcrzoex", 252,
  /*version*/       1, 0,
  /*title*/         "scan of IP range, for TCP port range (spoofed)",
  /*french title*/  "scan d'une plage IP, pour une plage de ports TCP (spoofe)",
  /*description*/   "",
  /*french desc.*/  "",
  /*usage*/         "device srceth srcip dstiprange dstportrange",
  /*french usage*/  "device ethsrc ipsrc plageipdst plageportdst",
  /*usage example*/ "eth0 aa:bb:cc:dd:ee:ff 1.2.3.4 5.6.7.0/24 1-120",
  /*fr. usage ex.*/ "eth0 aa:bb:cc:dd:ee:ff 1.2.3.4 5.6.7.0/24 1-120",
  /*author*/        "Laurent"
};

/*-------------------------------------------------------------*/
/* function declarations */
int lcrzoex_000252_main(int argc, char *argv[]);
void lcrzoex_000252_sig_received1(int i);
void lcrzoex_000252_sig_received2(int i);
int lcrzoex_000252_scaniptcp(lcrzo_device device,
			     lcrzo_etha clientetha,
			     lcrzo_ipl clientipl, 
			     lcrzo_iprange iprange, 
			     lcrzo_portrange portrange);
int lcrzoex_000252_sendsyn(lcrzo_device device,
			   lcrzo_etha clientetha,
			   lcrzo_ipl clientipl,
			   lcrzo_ipl serveriplinf,
			   lcrzo_ipl serveriplsup,
			   lcrzo_uint16 serverportinf,
			   lcrzo_uint16 serverportsup,
			   lcrzo_uint16 startingclientport,
			   lcrzo_uint32 startingclientseqnum);
int lcrzoex_000252_print(lcrzo_device device,
			 lcrzo_etha clientetha,
			 lcrzo_ipl clientipl,
			 lcrzo_ipl serveriplinf,
			 lcrzo_ipl serveriplsup,
			 lcrzo_uint16 serverportinf,
			 lcrzo_uint16 serverportsup,
			 lcrzo_uint16 startingclientport,
			 lcrzo_uint32 startingclientseqnum);
int lcrzoex_000252_printroutei(lcrzo_etha clientetha,
			       lcrzo_ipl clientipl,
			       lcrzo_ipl serveriplinf,
			       lcrzo_ipl serveriplsup,
			       lcrzo_uint16 serverportinf,
			       lcrzo_uint16 serverportsup,
			       lcrzo_uint16 startingclientport,
			       lcrzo_uint32 startingclientseqnum,
			       lcrzo_data sniffdata,
			       lcrzo_int32 sniffdatasize);
int lcrzoex_000252_printroutet(lcrzo_etha clientetha,
			       lcrzo_ipl clientipl,
			       lcrzo_ipl serveriplinf,
			       lcrzo_ipl serveriplsup,
			       lcrzo_uint16 serverportinf,
			       lcrzo_uint16 serverportsup,
			       lcrzo_uint16 startingclientport,
			       lcrzo_uint32 startingclientseqnum,
			       lcrzo_data sniffdata,
			       lcrzo_int32 sniffdatasize);

/*-------------------------------------------------------------*/
/* global variables */
int lcrzoex_000252_pid;
lcrzo_spoof lcrzoex_000252_spoof;
lcrzo_sniff lcrzoex_000252_sniff;

/*-------------------------------------------------------------*/
/* main function : you can rename it to "main", and then this file
   can be compiled alone. */
int lcrzoex_000252_main(int argc, char *argv[])
{ lcrzo_device device;
  lcrzo_etha clientetha;
  lcrzo_ipl clientipl;
  lcrzo_iprange iprange;
  lcrzo_portrange portrange;

  if ( argc<6 )
  { /* print usage */
    lcrzo_epr(lcrzoex_infos_print(lcrzoex_000252, argv[0]));
    return(LCRZO_ERR_BULCRZOEXBADUSAGE);
  }
  lcrzo_epr(lcrzo_device_init(argv[1], device));
  lcrzo_epr(lcrzo_etha_init_eths(argv[2], clientetha));
  lcrzo_epr(lcrzo_ipl_init_hs(argv[3], &clientipl));
  lcrzo_epr(lcrzo_string_init_text(argv[4], LCRZO_IPRANGE_MAXBYTES, iprange));
  lcrzo_epr(lcrzo_string_init_text(argv[5], LCRZO_PORTRANGE_MAXBYTES,
				   portrange));

  /* call the main function */
  lcrzo_epr(lcrzoex_000252_scaniptcp(device, clientetha, clientipl,
				     iprange, portrange));

  return (LCRZO_ERR_OK);
}

/*-------------------------------------------------------------*/
int lcrzoex_000252_scaniptcp(lcrzo_device device,
			     lcrzo_etha clientetha,
			     lcrzo_ipl clientipl, 
			     lcrzo_iprange iprange, 
			     lcrzo_portrange portrange)
{ lcrzo_uint16 startingclientport;
  lcrzo_uint32 startingclientseqnum;
  lcrzo_ipl ipinf, ipsup;
  lcrzo_uint16 portinf, portsup;
  
  /* determine starting values */
  lcrzo_epr(lcrzo_uint16_rand(5000, 30000, &startingclientport));
  lcrzo_epr(lcrzo_uint32_rand(0, 0xEFFFFFFFU, &startingclientseqnum));

  /* obtain starting and ending values */
  lcrzo_epr(lcrzo_ipl_init_iprange(iprange, &ipinf, &ipsup));
  lcrzo_epr(lcrzo_port_init_portrange(portrange, &portinf, &portsup));

  /* create child processes */
  lcrzoex_000252_pid=fork();
  if (lcrzoex_000252_pid<0) return(LCRZO_ERR_FUFORK);
  if (lcrzoex_000252_pid==0)
  { /* this process sends a SYN */
    /* in case we receive a signal */
    signal(SIGINT, lcrzoex_000252_sig_received1);
    lcrzo_epr(lcrzoex_000252_sendsyn(device, clientetha, clientipl,
				     ipinf, ipsup, portinf, portsup,
				     startingclientport,
				     startingclientseqnum));
  }
  else
  { /* this process sniff and prints sequence numbers */
    /* in case we receive a signal */
    signal(SIGINT, lcrzoex_000252_sig_received2);
    lcrzo_epr(lcrzoex_000252_print(device, clientetha, clientipl,
				   ipinf, ipsup, portinf, portsup,
				   startingclientport,
				   startingclientseqnum));
  }
  return (LCRZO_ERR_OK);
}

/*-------------------------------------------------------------*/
void lcrzoex_000252_sig_received1(int i)
{ int language;

  lcrzo_global_get_language(&language);
  if ( language==LCRZO_GLOBAL_LANG_FR )
    printf("Fermeture du spoof (recu : SIG%d)\n", i);
  else
    printf("Closing spoof (received : SIG%d)\n", i);
  lcrzo_spoof_close(&lcrzoex_000252_spoof);
  raise(SIGTERM);
}

/*-------------------------------------------------------------*/
void lcrzoex_000252_sig_received2(int i)
{ int status, language;

  lcrzo_global_get_language(&language);
  if ( language==LCRZO_GLOBAL_LANG_FR )
    printf("Fermeture du sniff (recu : SIG%d)\n", i);
  else
    printf("Closing sniff (received : SIG%d)\n", i);
  kill(lcrzoex_000252_pid, SIGINT);
  waitpid(lcrzoex_000252_pid, &status, 0);
  lcrzo_sniff_close(&lcrzoex_000252_sniff);
  raise(SIGTERM);
}

/*-------------------------------------------------------------*/
int lcrzoex_000252_sendsyn(lcrzo_device device,
			   lcrzo_etha clientetha,
			   lcrzo_ipl clientipl,
			   lcrzo_ipl serveriplinf,
			   lcrzo_ipl serveriplsup,
			   lcrzo_uint16 serverportinf,
			   lcrzo_uint16 serverportsup,
			   lcrzo_uint16 startingclientport,
			   lcrzo_uint32 startingclientseqnum)
{ lcrzo_hdrleth hdrleth;
  lcrzo_hdrlip hdrlip;
  lcrzo_hdrltcp hdrltcp;
  lcrzo_ipl serveripl;
  lcrzo_uint16 serverport;
  lcrzo_device deviceneeded;
  lcrzo_etha routeretha;
  int language, ret;

  /* init spoof */
  lcrzo_epr(lcrzo_spoof_init(&lcrzoex_000252_spoof));
  /*lcrzo_epr(lcrzo_spoof_set_printbeforesending(&lcrzoex_000252_spoof,
                                               LCRZO_TRUE));*/

  /* loop through each server */
  for ( serveripl=serveriplinf ; serveripl<=serveriplsup ; serveripl++ )
  { 
    /* obtain routeretha */
    ret=lcrzo_conf_route_to_host(serveripl, deviceneeded, NULL, routeretha,
				 NULL, NULL);
    if (ret!=LCRZO_ERR_OK)
    { lcrzo_ipl_print("", serveripl, " : ");
      lcrzo_err_print(ret);
      continue;
    }
    /* display a warning */
    if ( ! lcrzo_device_equal(device, deviceneeded) )
    { lcrzo_epr(lcrzo_global_get_language(&language));
      if ( language==LCRZO_GLOBAL_LANG_FR )
      { printf("Le device employe (%s) est different du device", device);
        printf(" necessaire (%s) pour joindre ", deviceneeded);
      }
      else
      { printf("The used device (%s) is different from the necessary", device);
        printf(" device (%s) to reach ", deviceneeded);
      }
      lcrzo_ipl_print("", serveripl, "\n");
    }

    /* initialize Ethernet header */
    lcrzo_epr(lcrzo_hdrleth_initdefault(&hdrleth));
    memcpy(hdrleth.src, clientetha, LCRZO_ETHA_MAXBYTES);
    memcpy(hdrleth.dst, routeretha, LCRZO_ETHA_MAXBYTES);

    /* initialize IP header */
    lcrzo_epr(lcrzo_hdrlip_initdefault(&hdrlip));
    hdrlip.saddr=clientipl;
    hdrlip.daddr=serveripl;
    hdrlip.id=startingclientport;

    /* initialize TCP header */
    lcrzo_epr(lcrzo_hdrltcp_initdefault(&hdrltcp));
    hdrltcp.dport=serverportinf;
    hdrltcp.syn=1;
    hdrltcp.sport=startingclientport;
    hdrltcp.seqnum=startingclientseqnum;

    /* spoof */
    for ( serverport=serverportinf ; serverport<=serverportsup ; serverport++ )
    { lcrzo_epr(lcrzo_spoof_ethipopttcpoptdata(&lcrzoex_000252_spoof,
					       device, hdrleth,
					       hdrlip, NULL, 0,
					       hdrltcp, NULL, 0,
					       NULL, 0));
      hdrlip.id++;
      hdrltcp.sport++;
      hdrltcp.dport++;
      hdrltcp.seqnum+=10;
      lcrzo_epr(lcrzo_time_sleep_usec(10000));
    }
  }

  /* close */
  lcrzo_epr(lcrzo_spoof_close(&lcrzoex_000252_spoof));  
  return(LCRZO_ERR_OK);
}

/*-------------------------------------------------------------*/
int lcrzoex_000252_print(lcrzo_device device,
			 lcrzo_etha clientetha,
			 lcrzo_ipl clientipl,
			 lcrzo_ipl serveriplinf,
			 lcrzo_ipl serveriplsup,
			 lcrzo_uint16 serverportinf,
			 lcrzo_uint16 serverportsup,
			 lcrzo_uint16 startingclientport,
			 lcrzo_uint32 startingclientseqnum)
{ lcrzo_uint32 mtu;
  lcrzo_data sniffdata;
  lcrzo_int32 sniffdatasize;
  int ret;

  /* init sniff */
  lcrzo_epr(lcrzo_mtu_init_device(device, &mtu));
  lcrzo_epr(lcrzo_sniff_init(device, mtu, "", &lcrzoex_000252_sniff));

  /* main loop */
  ret=lcrzo_sniff_nextm(&lcrzoex_000252_sniff, LCRZO_TRUE, 
			LCRZO_SNIFF_TYPE_NORMAL, &sniffdata, &sniffdatasize);
  while (ret==LCRZO_ERR_OK)
  { /* answer to requests */
    lcrzo_epr(lcrzo_virtual_answer_alive(sniffdata, sniffdatasize, 
                                         device, clientetha, clientipl));
    /* decode an icmp error, and print it */
    lcrzo_epr(lcrzoex_000252_printroutei(clientetha, clientipl,
					 serveriplinf, serveriplsup,
					 serverportinf, serverportsup,
					 startingclientport,
					 startingclientseqnum,
					 sniffdata, sniffdatasize));
    /* decode a tcp synack or reset, and print it */
    lcrzo_epr(lcrzoex_000252_printroutet(clientetha, clientipl,
					 serveriplinf, serveriplsup,
					 serverportinf, serverportsup,
					 startingclientport,
					 startingclientseqnum,
					 sniffdata, sniffdatasize));
    lcrzo_data_free(sniffdata);
    /* sniff next packet */
    ret=lcrzo_sniff_nextm(&lcrzoex_000252_sniff, LCRZO_TRUE,
			  LCRZO_SNIFF_TYPE_NORMAL, &sniffdata, &sniffdatasize);
  }

  /* close everything */
  raise(SIGINT);
  /*never reatched*/
  return(LCRZO_ERR_IEINTERNALERROR);
}

/*-------------------------------------------------------------*/
int lcrzoex_000252_printroutei(lcrzo_etha clientetha,
			       lcrzo_ipl clientipl,
			       lcrzo_ipl serveriplinf,
			       lcrzo_ipl serveriplsup,
			       lcrzo_uint16 serverportinf,
			       lcrzo_uint16 serverportsup,
			       lcrzo_uint16 startingclientport,
			       lcrzo_uint32 startingclientseqnum,
			       lcrzo_data sniffdata,
			       lcrzo_int32 sniffdatasize)
{ lcrzo_hdrleth hdrleth;
  lcrzo_hdrlip hdrlip, badhdrip;
  lcrzo_data baddata;
  lcrzo_string string;
  lcrzo_uint16 baddatasize, badtcpsport, badtcpdport;
  lcrzo_uint8 type, code;
  lcrzo_uint32 badtcpseqnum;
  int ret, indexidok, indexidko, indexsport, indexseqnum;

  /*lcrzo_epr(lcrzo_packet_print_eth(sniffdata, sniffdatasize,
				   LCRZO_PRINTPROFILE_ARRAYDUMP));*/

  /* try to decode icmp */
  ret=lcrzo_packet_decodem_ethipopticmperror(sniffdata, sniffdatasize,
					     &hdrleth, &hdrlip, NULL,NULL,
					     &type, &code, NULL, NULL,
					     &baddata, &baddatasize);
  if ( ret!=LCRZO_ERR_OK )
    return(LCRZO_ERR_OK);

  /* check if the addresses are the one we are looking for */
  if ( ! lcrzo_etha_equal(hdrleth.dst, clientetha) ||
       ! lcrzo_ipl_equal(hdrlip.daddr, clientipl) )
  { lcrzo_data_free(baddata);
    return(LCRZO_ERR_OK);
  }

  /* icmp should contain tcp */
  ret=lcrzo_ip64bits_decodem_iptcp(baddata, baddatasize,
				   &badhdrip, &badtcpsport, &badtcpdport,
				   &badtcpseqnum);
  lcrzo_data_free(baddata);
  if ( ret!=LCRZO_ERR_OK )
    return(LCRZO_ERR_OK);

  /* check various values in baddata */
  if ( badtcpdport<serverportinf ||
       badtcpdport>serverportsup ||
       badhdrip.saddr!=clientipl ||
       badhdrip.daddr<serveriplinf ||
       badhdrip.daddr>serveriplsup )
    return(LCRZO_ERR_OK);

  /* compute indexid from the id of the received packet */
  indexidok=badhdrip.id-startingclientport+1;
  /* this one is because there is a bug in the IP stack of BSD :
     id is reversed :( */
  indexidko= ( (badhdrip.id&0xFF)<<8 ) | ( (badhdrip.id&0xFF00)>>8 );
  indexidko=indexidko-startingclientport+1;
  /* compute indexsport from the sport of the received packet */
  indexsport=badtcpsport-startingclientport+1;
  /* compute indexseqnum from the seqnum of the received packet */
  indexseqnum=1+(badtcpseqnum-startingclientseqnum)/10;

  /* verify that indexes are equal */
  if ( (indexidok!=indexsport&&indexidko!=indexsport) ||
       indexsport!=indexseqnum )
    return(LCRZO_ERR_OK);

  /* display the error type and code */
  lcrzo_epr(lcrzo_ipl_print("", badhdrip.daddr, ""));
  printf(":%d - error : ", badtcpdport);
  lcrzo_epr(lcrzo_ipl_print("", hdrlip.saddr, " said "));
  lcrzo_epr(lcrzo_string_initm_icmptypecode(type, code, &string));
  printf(" : %s\n", string);
  lcrzo_string_free(string);

  return(LCRZO_ERR_OK);
}

/*-------------------------------------------------------------*/
int lcrzoex_000252_printroutet(lcrzo_etha clientetha,
			       lcrzo_ipl clientipl,
			       lcrzo_ipl serveriplinf,
			       lcrzo_ipl serveriplsup,
			       lcrzo_uint16 serverportinf,
			       lcrzo_uint16 serverportsup,
			       lcrzo_uint16 startingclientport,
			       lcrzo_uint32 startingclientseqnum,
			       lcrzo_data sniffdata,
			       lcrzo_int32 sniffdatasize)
{ lcrzo_hdrleth hdrleth;
  lcrzo_hdrlip hdrlip;
  lcrzo_hdrltcp hdrltcp;
  int ret, theindex1, theindex2, language;

  /*lcrzo_epr(lcrzo_packet_print_eth(sniffdata, sniffdatasize,
				   LCRZO_PRINTPROFILE_ARRAYDUMP));*/

  /* try to decode it */
  ret=lcrzo_packet_decodem_ethipopttcpoptdata(sniffdata, sniffdatasize,
					      &hdrleth, &hdrlip, NULL, NULL,
					      &hdrltcp, NULL, NULL,
					      NULL, NULL);
  if ( ret!=LCRZO_ERR_OK )
    return(LCRZO_ERR_OK);

  /* check if the addresses are the one we are looking for */
  if ( ! lcrzo_etha_equal(hdrleth.dst, clientetha) || 
       hdrlip.saddr<serveriplinf || hdrlip.saddr>serveriplsup || 
       ! lcrzo_ipl_equal(hdrlip.daddr, clientipl) )
    return(LCRZO_ERR_OK);

  /* check server port */
  if ( hdrltcp.sport<serverportinf || hdrltcp.sport>serverportsup )
    return(LCRZO_ERR_OK);

  /* compute the theindex from the received packet */
  theindex1=hdrltcp.dport-startingclientport+1;
  theindex2=1+(hdrltcp.acknum-1-startingclientseqnum)/10;
  if ( theindex1!=theindex2 )
    return(LCRZO_ERR_OK);

  /* if we received synack or reset, destination is reatched */
  if ( ( hdrltcp.syn && hdrltcp.ack ) || hdrltcp.rst )
  { lcrzo_epr(lcrzo_ipl_print("", hdrlip.saddr, ""));
    printf(":%d", hdrltcp.sport);
    lcrzo_epr(lcrzo_global_get_language(&language));
    if ( language==LCRZO_GLOBAL_LANG_FR )
    { if (hdrltcp.rst) printf(" - ferme\n");
      else printf(" - ouvert\n");
    }
    else
    { if (hdrltcp.rst) printf(" - closed\n");
      else printf(" - open\n");
    }
  }

  return(LCRZO_ERR_OK);
}
