
/* include the library headers */
#include <lcrzo.h>
#include "lcrzoex.h"
#include "shared/shared_misc.h"

/*-------------------------------------------------------------*/
/* set information about this example */
lcrzoex_infos lcrzoex_000254=
{ /*reference*/     "lcrzoex", 254,
  /*version*/       1, 0,
  /*title*/         "scan of IP range, for UDP port range (spoofed)",
  /*french title*/  "scan d'une plage IP, pour une plage de ports UDP (spoofe)",
  /*description*/   "",
  /*french desc.*/  "",
  /*usage*/         "device srceth srcip dstiprange dstportrange",
  /*french usage*/  "device ethsrc ipsrc plageipdst plageportdst",
  /*usage example*/ "eth0 aa:bb:cc:dd:ee:ff 1.2.3.4 5.6.7.0/24 1-120",
  /*fr. usage ex.*/ "eth0 aa:bb:cc:dd:ee:ff 1.2.3.4 5.6.7.0/24 1-120",
  /*author*/        "Laurent"
};

/*-------------------------------------------------------------*/
/* function declarations */
int lcrzoex_000254_main(int argc, char *argv[]);
void lcrzoex_000254_sig_received1(int i);
void lcrzoex_000254_sig_received2(int i);
int lcrzoex_000254_scanipudp(lcrzo_device device,
			     lcrzo_etha clientetha,
			     lcrzo_ipl clientipl, 
			     lcrzo_iprange iprange, 
			     lcrzo_portrange portrange);
int lcrzoex_000254_sendudp(lcrzo_device device,
			   lcrzo_etha clientetha,
			   lcrzo_ipl clientipl,
			   lcrzo_ipl serveriplinf,
			   lcrzo_ipl serveriplsup,
			   lcrzo_uint16 serverportinf,
			   lcrzo_uint16 serverportsup,
			   lcrzo_uint16 startingclientport);
int lcrzoex_000254_print(lcrzo_device device,
			 lcrzo_etha clientetha,
			 lcrzo_ipl clientipl,
			 lcrzo_ipl serveriplinf,
			 lcrzo_ipl serveriplsup,
			 lcrzo_uint16 serverportinf,
			 lcrzo_uint16 serverportsup,
			 lcrzo_uint16 startingclientport);
int lcrzoex_000254_printroutei(lcrzo_etha clientetha,
			       lcrzo_ipl clientipl,
			       lcrzo_ipl serveriplinf,
			       lcrzo_ipl serveriplsup,
			       lcrzo_uint16 serverportinf,
			       lcrzo_uint16 serverportsup,
			       lcrzo_uint16 startingclientport,
			       lcrzo_data sniffdata,
			       lcrzo_int32 sniffdatasize);

/*-------------------------------------------------------------*/
/* global variables */
int lcrzoex_000254_pid;
lcrzo_spoof lcrzoex_000254_spoof;
lcrzo_sniff lcrzoex_000254_sniff;

/*-------------------------------------------------------------*/
/* main function : you can rename it to "main", and then this file
   can be compiled alone. */
int lcrzoex_000254_main(int argc, char *argv[])
{ lcrzo_device device;
  lcrzo_etha clientetha;
  lcrzo_ipl clientipl;
  lcrzo_iprange iprange;
  lcrzo_portrange portrange;

  if ( argc<6 )
  { /* print usage */
    lcrzo_epr(lcrzoex_infos_print(lcrzoex_000254, argv[0]));
    return(LCRZO_ERR_BULCRZOEXBADUSAGE);
  }
  lcrzo_epr(lcrzo_device_init(argv[1], device));
  lcrzo_epr(lcrzo_etha_init_eths(argv[2], clientetha));
  lcrzo_epr(lcrzo_ipl_init_hs(argv[3], &clientipl));
  lcrzo_epr(lcrzo_string_init_text(argv[4], LCRZO_IPRANGE_MAXBYTES, iprange));
  lcrzo_epr(lcrzo_string_init_text(argv[5], LCRZO_PORTRANGE_MAXBYTES,
				   portrange));

  /* call the main function */
  lcrzo_epr(lcrzoex_000254_scanipudp(device, clientetha, clientipl,
				     iprange, portrange));

  return (LCRZO_ERR_OK);
}

/*-------------------------------------------------------------*/
int lcrzoex_000254_scanipudp(lcrzo_device device,
			     lcrzo_etha clientetha,
			     lcrzo_ipl clientipl, 
			     lcrzo_iprange iprange, 
			     lcrzo_portrange portrange)
{ lcrzo_uint16 startingclientport;
  lcrzo_ipl ipinf, ipsup;
  lcrzo_uint16 portinf, portsup;
  
  /* determine starting values */
  lcrzo_epr(lcrzo_uint16_rand(5000, 30000, &startingclientport));

  /* obtain starting and ending values */
  lcrzo_epr(lcrzo_ipl_init_iprange(iprange, &ipinf, &ipsup));
  lcrzo_epr(lcrzo_port_init_portrange(portrange, &portinf, &portsup));

  /* create child processes */
  lcrzoex_000254_pid=fork();
  if (lcrzoex_000254_pid<0) return(LCRZO_ERR_FUFORK);
  if (lcrzoex_000254_pid==0)
  { /* this process sends a packet */
    /* in case we receive a signal */
    signal(SIGINT, lcrzoex_000254_sig_received1);
    lcrzo_epr(lcrzoex_000254_sendudp(device, clientetha, clientipl,
				     ipinf, ipsup, portinf, portsup,
				     startingclientport));
  }
  else
  { /* this process sniff and prints sequence numbers */
    /* in case we receive a signal */
    signal(SIGINT, lcrzoex_000254_sig_received2);
    lcrzo_epr(lcrzoex_000254_print(device, clientetha, clientipl,
				   ipinf, ipsup, portinf, portsup,
				   startingclientport));
  }
  return (LCRZO_ERR_OK);
}

/*-------------------------------------------------------------*/
void lcrzoex_000254_sig_received1(int i)
{ int language;

  lcrzo_global_get_language(&language);
  if ( language==LCRZO_GLOBAL_LANG_FR )
    printf("Fermeture du spoof (recu : SIG%d)\n", i);
  else
    printf("Closing spoof (received : SIG%d)\n", i);
  lcrzo_spoof_close(&lcrzoex_000254_spoof);
  raise(SIGTERM);
}

/*-------------------------------------------------------------*/
void lcrzoex_000254_sig_received2(int i)
{ int status, language;

  lcrzo_global_get_language(&language);
  if ( language==LCRZO_GLOBAL_LANG_FR )
    printf("Fermeture du sniff (recu : SIG%d)\n", i);
  else
    printf("Closing sniff (received : SIG%d)\n", i);
  kill(lcrzoex_000254_pid, SIGINT);
  waitpid(lcrzoex_000254_pid, &status, 0);
  lcrzo_sniff_close(&lcrzoex_000254_sniff);
  raise(SIGTERM);
}

/*-------------------------------------------------------------*/
int lcrzoex_000254_sendudp(lcrzo_device device,
			   lcrzo_etha clientetha,
			   lcrzo_ipl clientipl,
			   lcrzo_ipl serveriplinf,
			   lcrzo_ipl serveriplsup,
			   lcrzo_uint16 serverportinf,
			   lcrzo_uint16 serverportsup,
			   lcrzo_uint16 startingclientport)
{ lcrzo_hdrleth hdrleth;
  lcrzo_hdrlip hdrlip;
  lcrzo_hdrludp hdrludp;
  lcrzo_ipl serveripl;
  lcrzo_uint16 serverport;
  lcrzo_device deviceneeded;
  lcrzo_etha routeretha;
  int language, ret;

  /* init spoof */
  lcrzo_epr(lcrzo_spoof_init(&lcrzoex_000254_spoof));
  /*lcrzo_epr(lcrzo_spoof_set_printbeforesending(&lcrzoex_000254_spoof,
                                               LCRZO_TRUE));*/

  /* loop through each server */
  for ( serveripl=serveriplinf ; serveripl<=serveriplsup ; serveripl++ )
  { 
    /* obtain routeretha */
    ret=lcrzo_conf_route_to_host(serveripl, deviceneeded, NULL, routeretha,
				 NULL, NULL);
    if (ret!=LCRZO_ERR_OK)
    { lcrzo_ipl_print("", serveripl, " : ");
      lcrzo_err_print(ret);
      continue;
    }
    /* display a warning */
    if ( ! lcrzo_device_equal(device, deviceneeded) )
    { lcrzo_epr(lcrzo_global_get_language(&language));
      if ( language==LCRZO_GLOBAL_LANG_FR )
      { printf("Le device employe (%s) est different du device", device);
        printf(" necessaire (%s) pour joindre ", deviceneeded);
      }
      else
      { printf("The used device (%s) is different from the necessary", device);
        printf(" device (%s) to reach ", deviceneeded);
      }
      lcrzo_ipl_print("", serveripl, "\n");
    }

    /* initialize Ethernet header */
    lcrzo_epr(lcrzo_hdrleth_initdefault(&hdrleth));
    memcpy(hdrleth.src, clientetha, LCRZO_ETHA_MAXBYTES);
    memcpy(hdrleth.dst, routeretha, LCRZO_ETHA_MAXBYTES);

    /* initialize IP header */
    lcrzo_epr(lcrzo_hdrlip_initdefault(&hdrlip));
    hdrlip.saddr=clientipl;
    hdrlip.daddr=serveripl;
    hdrlip.id=startingclientport;

    /* initialize UDP header */
    lcrzo_epr(lcrzo_hdrludp_initdefault(&hdrludp));
    hdrludp.dport=serverportinf;
    hdrludp.sport=startingclientport;

    /* spoof */
    for ( serverport=serverportinf ; serverport<=serverportsup ; serverport++ )
    { lcrzo_epr(lcrzo_spoof_ethipoptudpdata(&lcrzoex_000254_spoof,
					    device, hdrleth,
					    hdrlip, NULL, 0,
					    hdrludp, NULL, 0));
      hdrlip.id++;
      hdrludp.sport++;
      hdrludp.dport++;
      lcrzo_epr(lcrzo_time_sleep_sec(1));
    }
  }

  /* close */
  lcrzo_epr(lcrzo_spoof_close(&lcrzoex_000254_spoof));  
  return(LCRZO_ERR_OK);
}

/*-------------------------------------------------------------*/
int lcrzoex_000254_print(lcrzo_device device,
			 lcrzo_etha clientetha,
			 lcrzo_ipl clientipl,
			 lcrzo_ipl serveriplinf,
			 lcrzo_ipl serveriplsup,
			 lcrzo_uint16 serverportinf,
			 lcrzo_uint16 serverportsup,
			 lcrzo_uint16 startingclientport)
{ lcrzo_uint32 mtu;
  lcrzo_data sniffdata;
  lcrzo_int32 sniffdatasize;
  int ret;

  /* init sniff */
  lcrzo_epr(lcrzo_mtu_init_device(device, &mtu));
  lcrzo_epr(lcrzo_sniff_init(device, mtu, "", &lcrzoex_000254_sniff));

  /* main loop */
  ret=lcrzo_sniff_nextm(&lcrzoex_000254_sniff, LCRZO_TRUE, 
			LCRZO_SNIFF_TYPE_NORMAL, &sniffdata, &sniffdatasize);
  while (ret==LCRZO_ERR_OK)
  { /* answer to requests */
    lcrzo_epr(lcrzo_virtual_answer_alive(sniffdata, sniffdatasize, 
                                         device, clientetha, clientipl));
    /* decode an icmp error, and print it */
    lcrzo_epr(lcrzoex_000254_printroutei(clientetha, clientipl,
					 serveriplinf, serveriplsup,
					 serverportinf, serverportsup,
					 startingclientport,
					 sniffdata, sniffdatasize));
    lcrzo_data_free(sniffdata);
    /* sniff next packet */
    ret=lcrzo_sniff_nextm(&lcrzoex_000254_sniff, LCRZO_TRUE,
			  LCRZO_SNIFF_TYPE_NORMAL, &sniffdata, &sniffdatasize);
  }

  /* close everything */
  raise(SIGINT);
  /*never reatched*/
  return(LCRZO_ERR_IEINTERNALERROR);
}

/*-------------------------------------------------------------*/
int lcrzoex_000254_printroutei(lcrzo_etha clientetha,
			       lcrzo_ipl clientipl,
			       lcrzo_ipl serveriplinf,
			       lcrzo_ipl serveriplsup,
			       lcrzo_uint16 serverportinf,
			       lcrzo_uint16 serverportsup,
			       lcrzo_uint16 startingclientport,
			       lcrzo_data sniffdata,
			       lcrzo_int32 sniffdatasize)
{ lcrzo_hdrleth hdrleth;
  lcrzo_hdrlip hdrlip, badhdrip;
  lcrzo_data baddata;
  lcrzo_string string;
  lcrzo_uint16 baddatasize, badudpsport, badudpdport;
  lcrzo_uint8 type, code;
  int ret, indexidok, indexidko, indexsport, language;

  /*lcrzo_epr(lcrzo_packet_print_eth(sniffdata, sniffdatasize,
				   LCRZO_PRINTPROFILE_ARRAYDUMP));*/

  /* try to decode icmp */
  ret=lcrzo_packet_decodem_ethipopticmperror(sniffdata, sniffdatasize,
					     &hdrleth, &hdrlip, NULL,NULL,
					     &type, &code, NULL, NULL,
					     &baddata, &baddatasize);
  if ( ret!=LCRZO_ERR_OK )
    return(LCRZO_ERR_OK);

  /* check if the addresses are the one we are looking for */
  if ( ! lcrzo_etha_equal(hdrleth.dst, clientetha) ||
       ! lcrzo_ipl_equal(hdrlip.daddr, clientipl) )
  { lcrzo_data_free(baddata);
    return(LCRZO_ERR_OK);
  }

  /* icmp should contain udp */
  ret=lcrzo_ip64bits_decodem_ipudp(baddata, baddatasize,
                                   &badhdrip, &badudpsport, &badudpdport,
                                   NULL, NULL);
  lcrzo_data_free(baddata);
  if ( ret!=LCRZO_ERR_OK )
    return(LCRZO_ERR_OK);

  /* check various values in baddata */
  if ( badudpdport<serverportinf ||
       badudpdport>serverportsup ||
       badhdrip.saddr!=clientipl ||
       badhdrip.daddr<serveriplinf ||
       badhdrip.daddr>serveriplsup )
    return(LCRZO_ERR_OK);

  /* compute indexid from the id of the received packet */
  indexidok=badhdrip.id-startingclientport+1;
  /* this one is because there is a bug in the IP stack of BSD :
     id is reversed :( */
  indexidko= ( (badhdrip.id&0xFF)<<8 ) | ( (badhdrip.id&0xFF00)>>8 );
  indexidko=indexidko-startingclientport+1;
  /* compute indexsport from the sport of the received packet */
  indexsport=badudpsport-startingclientport+1;

  /* verify that indexes are equal */
  if ( (indexidok!=indexsport&&indexidko!=indexsport) )
    return(LCRZO_ERR_OK);

  /* display the error type and code */
  lcrzo_epr(lcrzo_global_get_language(&language));
  lcrzo_epr(lcrzo_ipl_print("", badhdrip.daddr, ""));
  printf(":%d", badudpdport);
  if ( type==LCRZO_ICMPTYPE_DESTUNREACH &&
       code == LCRZO_ICMPCODE_DESTUNREACH_PORT )
  { if ( language==LCRZO_GLOBAL_LANG_FR )
      printf(" - ferme\n");
    else
      printf(" - closed\n");
  }
  else
  { if ( language==LCRZO_GLOBAL_LANG_FR )
      printf(" - error : ");
    else
      printf(" - erreur : ");
    lcrzo_epr(lcrzo_ipl_print("", hdrlip.saddr, ""));
    if ( language==LCRZO_GLOBAL_LANG_FR )
      printf(" a dit : ");
    else
      printf(" said : ");
    lcrzo_epr(lcrzo_string_initm_icmptypecode(type, code, &string));
    printf(" : %s\n", string);
    lcrzo_string_free(string);
  }
  return(LCRZO_ERR_OK);
}

