
/* include the library headers */
#include <lcrzo.h>
#include "lcrzoex.h"
#include "shared/shared_misc.h"

/*-------------------------------------------------------------*/
/* set information about this example */
lcrzoex_infos lcrzoex_000256=
{ /*reference*/     "lcrzoex", 256,
  /*version*/       1, 0,
  /*title*/         "scan of IP range, for ICMP ping (spoofed)",
  /*french title*/  "scan d'une plage IP, pour des ping ICMP (spoofe)",
  /*description*/   "",
  /*french desc.*/  "",
  /*usage*/         "device srceth srcip dstiprange",
  /*french usage*/  "device ethsrc ipsrc plageipdst",
  /*usage example*/ "eth0 aa:bb:cc:dd:ee:ff 1.2.3.4 5.6.7.0/24",
  /*fr. usage ex.*/ "eth0 aa:bb:cc:dd:ee:ff 1.2.3.4 5.6.7.0/24",
  /*author*/        "Laurent"
};

/*-------------------------------------------------------------*/
/* function declarations */
int lcrzoex_000256_main(int argc, char *argv[]);
void lcrzoex_000256_sig_received1(int i);
void lcrzoex_000256_sig_received2(int i);
int lcrzoex_000256_scanipicmp(lcrzo_device device,
			     lcrzo_etha clientetha,
			     lcrzo_ipl clientipl, 
			     lcrzo_iprange iprange);
int lcrzoex_000256_sendicmp(lcrzo_device device,
			    lcrzo_etha clientetha,
			    lcrzo_ipl clientipl,
			    lcrzo_ipl serveriplinf,
			    lcrzo_ipl serveriplsup,
			    lcrzo_uint16 startingid);
int lcrzoex_000256_print(lcrzo_device device,
			 lcrzo_etha clientetha,
			 lcrzo_ipl clientipl,
			 lcrzo_ipl serveriplinf,
			 lcrzo_ipl serveriplsup,
			 lcrzo_uint16 startingid);
int lcrzoex_000256_printroutei(lcrzo_etha clientetha,
			       lcrzo_ipl clientipl,
			       lcrzo_ipl serveriplinf,
			       lcrzo_ipl serveriplsup,
			       lcrzo_uint16 startingid,
			       lcrzo_data sniffdata,
			       lcrzo_int32 sniffdatasize);
int lcrzoex_000256_printroutet(lcrzo_etha clientetha,
			       lcrzo_ipl clientipl,
			       lcrzo_ipl serveriplinf,
			       lcrzo_ipl serveriplsup,
			       lcrzo_uint16 startingid,
			       lcrzo_data sniffdata,
			       lcrzo_int32 sniffdatasize);

/*-------------------------------------------------------------*/
/* global variables */
int lcrzoex_000256_pid;
lcrzo_spoof lcrzoex_000256_spoof;
lcrzo_sniff lcrzoex_000256_sniff;

/*-------------------------------------------------------------*/
/* main function : you can rename it to "main", and then this file
   can be compiled alone. */
int lcrzoex_000256_main(int argc, char *argv[])
{ lcrzo_device device;
  lcrzo_etha clientetha;
  lcrzo_ipl clientipl;
  lcrzo_iprange iprange;

  if ( argc<5 )
  { /* print usage */
    lcrzo_epr(lcrzoex_infos_print(lcrzoex_000256, argv[0]));
    return(LCRZO_ERR_BULCRZOEXBADUSAGE);
  }
  lcrzo_epr(lcrzo_device_init(argv[1], device));
  lcrzo_epr(lcrzo_etha_init_eths(argv[2], clientetha));
  lcrzo_epr(lcrzo_ipl_init_hs(argv[3], &clientipl));
  lcrzo_epr(lcrzo_string_init_text(argv[4], LCRZO_IPRANGE_MAXBYTES, iprange));

  /* call the main function */
  lcrzo_epr(lcrzoex_000256_scanipicmp(device, clientetha, clientipl, iprange));

  return (LCRZO_ERR_OK);
}

/*-------------------------------------------------------------*/
int lcrzoex_000256_scanipicmp(lcrzo_device device,
			      lcrzo_etha clientetha,
			      lcrzo_ipl clientipl, 
			      lcrzo_iprange iprange)
{ lcrzo_uint16 startingid;
  lcrzo_ipl ipinf, ipsup;
  
  /* determine starting values */
  lcrzo_epr(lcrzo_uint16_rand(5000, 30000, &startingid));

  /* obtain starting and ending values */
  lcrzo_epr(lcrzo_ipl_init_iprange(iprange, &ipinf, &ipsup));

  /* create child processes */
  lcrzoex_000256_pid=fork();
  if (lcrzoex_000256_pid<0) return(LCRZO_ERR_FUFORK);
  if (lcrzoex_000256_pid==0)
  { /* this process sends a echo request */
    /* in case we receive a signal */
    signal(SIGINT, lcrzoex_000256_sig_received1);
    lcrzo_epr(lcrzoex_000256_sendicmp(device, clientetha, clientipl,
				      ipinf, ipsup, startingid));
  }
  else
  { /* this process sniff and prints sequence numbers */
    /* in case we receive a signal */
    signal(SIGINT, lcrzoex_000256_sig_received2);
    lcrzo_epr(lcrzoex_000256_print(device, clientetha, clientipl,
				   ipinf, ipsup, startingid));
  }
  return (LCRZO_ERR_OK);
}

/*-------------------------------------------------------------*/
void lcrzoex_000256_sig_received1(int i)
{ int language;

  lcrzo_global_get_language(&language);
  if ( language==LCRZO_GLOBAL_LANG_FR )
    printf("Fermeture du spoof (recu : SIG%d)\n", i);
  else
    printf("Closing spoof (received : SIG%d)\n", i);
  lcrzo_spoof_close(&lcrzoex_000256_spoof);
  raise(SIGTERM);
}

/*-------------------------------------------------------------*/
void lcrzoex_000256_sig_received2(int i)
{ int status, language;

  lcrzo_global_get_language(&language);
  if ( language==LCRZO_GLOBAL_LANG_FR )
    printf("Fermeture du sniff (recu : SIG%d)\n", i);
  else
    printf("Closing sniff (received : SIG%d)\n", i);
  kill(lcrzoex_000256_pid, SIGINT);
  waitpid(lcrzoex_000256_pid, &status, 0);
  lcrzo_sniff_close(&lcrzoex_000256_sniff);
  raise(SIGTERM);
}

/*-------------------------------------------------------------*/
int lcrzoex_000256_sendicmp(lcrzo_device device,
			    lcrzo_etha clientetha,
			    lcrzo_ipl clientipl,
			    lcrzo_ipl serveriplinf,
			    lcrzo_ipl serveriplsup,
			    lcrzo_uint16 startingid)
{ lcrzo_hdrleth hdrleth;
  lcrzo_hdrlip hdrlip;
  lcrzo_data data;
  lcrzo_int32 datasize; 
  lcrzo_ipl serveripl;
  lcrzo_device deviceneeded;
  lcrzo_etha routeretha;
  int language, ret;

  /* init spoof */
  lcrzo_epr(lcrzo_spoof_init(&lcrzoex_000256_spoof));
  /*lcrzo_epr(lcrzo_spoof_set_printbeforesending(&lcrzoex_000256_spoof,
                                               LCRZO_TRUE));*/

  /* loop through each server */
  for ( serveripl=serveriplinf ; serveripl<=serveriplsup ; serveripl++ )
  { 
    /* obtain routeretha */
    ret=lcrzo_conf_route_to_host(serveripl, deviceneeded, NULL, routeretha,
				 NULL, NULL);
    if (ret!=LCRZO_ERR_OK)
    { lcrzo_ipl_print("", serveripl, " : ");
      lcrzo_err_print(ret);
      continue;
    }
    /* display a warning */
    if ( ! lcrzo_device_equal(device, deviceneeded) )
    { lcrzo_epr(lcrzo_global_get_language(&language));
      if ( language==LCRZO_GLOBAL_LANG_FR )
      { printf("Le device employe (%s) est different du device", device);
        printf(" necessaire (%s) pour joindre ", deviceneeded);
      }
      else
      { printf("The used device (%s) is different from the necessary", device);
        printf(" device (%s) to reach ", deviceneeded);
      }
      lcrzo_ipl_print("", serveripl, "\n");
    }

    /* initialize Ethernet header */
    lcrzo_epr(lcrzo_hdrleth_initdefault(&hdrleth));
    memcpy(hdrleth.src, clientetha, LCRZO_ETHA_MAXBYTES);
    memcpy(hdrleth.dst, routeretha, LCRZO_ETHA_MAXBYTES);

    /* initialize IP header */
    lcrzo_epr(lcrzo_hdrlip_initdefault(&hdrlip));
    hdrlip.saddr=clientipl;
    hdrlip.daddr=serveripl;
    hdrlip.id=startingid;

    /* spoof */
    lcrzo_epr(lcrzo_packet_initm_ethipopticmpecho(hdrleth,
                                                  hdrlip, NULL, 0,
                                                  LCRZO_ICMPTYPE_ECHOREQ,
                                                  startingid, startingid,
                                                  "ICMP scan", 9,
                                                  &data, &datasize));
    lcrzo_epr(lcrzo_spoof_eth(&lcrzoex_000256_spoof, device, data, datasize));
    lcrzo_epr(lcrzo_data_free(data));
    hdrlip.id++;
    startingid++;
    lcrzo_epr(lcrzo_time_sleep_usec(10000));
  }

  /* close */
  lcrzo_epr(lcrzo_spoof_close(&lcrzoex_000256_spoof));  
  return(LCRZO_ERR_OK);
}

/*-------------------------------------------------------------*/
int lcrzoex_000256_print(lcrzo_device device,
			 lcrzo_etha clientetha,
			 lcrzo_ipl clientipl,
			 lcrzo_ipl serveriplinf,
			 lcrzo_ipl serveriplsup,
			 lcrzo_uint16 startingid)
{ lcrzo_uint32 mtu;
  lcrzo_data sniffdata;
  lcrzo_int32 sniffdatasize;
  int ret;

  /* init sniff */
  lcrzo_epr(lcrzo_mtu_init_device(device, &mtu));
  lcrzo_epr(lcrzo_sniff_init(device, mtu, "", &lcrzoex_000256_sniff));

  /* main loop */
  ret=lcrzo_sniff_nextm(&lcrzoex_000256_sniff, LCRZO_TRUE, 
			LCRZO_SNIFF_TYPE_NORMAL, &sniffdata, &sniffdatasize);
  while (ret==LCRZO_ERR_OK)
  { /* answer to requests */
    lcrzo_epr(lcrzo_virtual_answer_alive(sniffdata, sniffdatasize, 
                                         device, clientetha, clientipl));
    /* decode an icmp error, and print it */
    lcrzo_epr(lcrzoex_000256_printroutei(clientetha, clientipl,
					 serveriplinf, serveriplsup,
					 startingid,
					 sniffdata, sniffdatasize));
    /* decode a icmp synack or reset, and print it */
    lcrzo_epr(lcrzoex_000256_printroutet(clientetha, clientipl,
					 serveriplinf, serveriplsup,
					 startingid,
					 sniffdata, sniffdatasize));
    lcrzo_data_free(sniffdata);
    /* sniff next packet */
    ret=lcrzo_sniff_nextm(&lcrzoex_000256_sniff, LCRZO_TRUE,
			  LCRZO_SNIFF_TYPE_NORMAL, &sniffdata, &sniffdatasize);
  }

  /* close everything */
  raise(SIGINT);
  /*never reatched*/
  return(LCRZO_ERR_IEINTERNALERROR);
}

/*-------------------------------------------------------------*/
int lcrzoex_000256_printroutei(lcrzo_etha clientetha,
			       lcrzo_ipl clientipl,
			       lcrzo_ipl serveriplinf,
			       lcrzo_ipl serveriplsup,
			       lcrzo_uint16 startingid,
			       lcrzo_data sniffdata,
			       lcrzo_int32 sniffdatasize)
{ lcrzo_hdrleth hdrleth;
  lcrzo_hdrlip hdrlip, badhdrip;
  lcrzo_data baddata;
  lcrzo_string string;
  lcrzo_uint16 baddatasize;
  lcrzo_uint8 type, code, badicmptype, badicmpcode, b4, b5, b6, b7;
  int ret, indexidok, indexidko, indexechoid, indexechoseqnum;

  /*lcrzo_epr(lcrzo_packet_print_eth(sniffdata, sniffdatasize,
				   LCRZO_PRINTPROFILE_ARRAYDUMP));*/

  /* try to decode icmp */
  ret=lcrzo_packet_decodem_ethipopticmperror(sniffdata, sniffdatasize,
					     &hdrleth, &hdrlip, NULL,NULL,
					     &type, &code, NULL, NULL,
					     &baddata, &baddatasize);
  if ( ret!=LCRZO_ERR_OK )
    return(LCRZO_ERR_OK);

  /* check if the addresses are the one we are looking for */
  if ( ! lcrzo_etha_equal(hdrleth.dst, clientetha) ||
       ! lcrzo_ipl_equal(hdrlip.daddr, clientipl) )
  { lcrzo_data_free(baddata);
    return(LCRZO_ERR_OK);
  }

  /* icmp should contain icmp */
  ret=lcrzo_ip64bits_decodem_ipicmp(baddata, baddatasize,
                                    &badhdrip, &badicmptype, &badicmpcode,
                                    NULL, &b4, &b5, &b6, &b7);
  lcrzo_data_free(baddata);
  if ( ret!=LCRZO_ERR_OK )
    return(LCRZO_ERR_OK);

  /* check various values in baddata */
  if ( badicmptype!=LCRZO_ICMPTYPE_ECHOREQ ||
       badicmpcode!=0 ||
        badhdrip.saddr!=clientipl ||
       badhdrip.daddr<serveriplinf ||
       badhdrip.daddr>serveriplsup )
    return(LCRZO_ERR_OK);

  /* compute indexid from the id of the received packet */
  indexidok=badhdrip.id-startingid+1;
  /* this one is because there is a bug in the IP stack of BSD :
     id is reversed :( */
  indexidko= ( (badhdrip.id&0xFF)<<8 ) | ( (badhdrip.id&0xFF00)>>8 );
  indexidko=indexidko-startingid+1;
  /* compute indexechoid from the id of the received packet */
  indexechoid=((b4<<8)|b5)-startingid+1;
  /* compute indexechoseqnum from the seqnum of the received packet */
  indexechoseqnum=((b6<<8)|b7)-startingid+1;

  /* verify that indexes are equal */
  if ( (indexidok!=indexechoid&&indexidko!=indexechoid) ||
       indexechoid!=indexechoseqnum )
    return(LCRZO_ERR_OK);

  /* display the error type and code */
  lcrzo_epr(lcrzo_ipl_print("", badhdrip.daddr, " - error : "));
  lcrzo_epr(lcrzo_ipl_print("", hdrlip.saddr, " said "));
  lcrzo_epr(lcrzo_string_initm_icmptypecode(type, code, &string));
  printf(" : %s\n", string);
  lcrzo_string_free(string);

  return(LCRZO_ERR_OK);
}

/*-------------------------------------------------------------*/
int lcrzoex_000256_printroutet(lcrzo_etha clientetha,
			       lcrzo_ipl clientipl,
			       lcrzo_ipl serveriplinf,
			       lcrzo_ipl serveriplsup,
			       lcrzo_uint16 startingid,
			       lcrzo_data sniffdata,
			       lcrzo_int32 sniffdatasize)
{ lcrzo_hdrleth hdrleth;
  lcrzo_hdrlip hdrlip;
  lcrzo_uint8 type;
  lcrzo_uint16 id, seqnum;
  int ret, theindex1, theindex2, language;

  /*lcrzo_epr(lcrzo_packet_print_eth(sniffdata, sniffdatasize,
				   LCRZO_PRINTPROFILE_ARRAYDUMP));*/

  /* try to decode it */
  ret=lcrzo_packet_decodem_ethipopticmpecho(sniffdata, sniffdatasize,
                                            &hdrleth, &hdrlip, NULL, NULL,
                                            &type, &id, &seqnum, NULL, NULL);
  if ( ret!=LCRZO_ERR_OK )
    return(LCRZO_ERR_OK);

  /* check if the addresses are the one we are looking for */
  if ( ! lcrzo_etha_equal(hdrleth.dst, clientetha) || 
       hdrlip.saddr<serveriplinf || hdrlip.saddr>serveriplsup || 
       ! lcrzo_ipl_equal(hdrlip.daddr, clientipl) )
    return(LCRZO_ERR_OK);

  /* check type */
  if ( type!=LCRZO_ICMPTYPE_ECHOREP )
    return(LCRZO_ERR_OK);

  /* compute the theindex from the received packet */
  theindex1=id-startingid+1;
  theindex2=seqnum-startingid+1;
  if ( theindex1!=theindex2 )
    return(LCRZO_ERR_OK);

  /* destination is reatched */
  lcrzo_epr(lcrzo_ipl_print("", hdrlip.saddr, ""));
  lcrzo_epr(lcrzo_global_get_language(&language));
  if ( language==LCRZO_GLOBAL_LANG_FR )
  { printf(" - atteint\n");
  }
  else
  { printf(" - reatched\n");
  }

  return(LCRZO_ERR_OK);
}





