#include <lcrzo.h>
#include "../lcrzoex.h"
#include "shared_misc.h"
#include "shared_nntp.h"
#include "shared_smtp.h"


/*-------------------------------------------------------------*/
int shared_nntp_appendm_line_newsgroups(lcrzo_conststring newsgroup,
					lcrzo_string *poutput)
{ lcrzo_epr(lcrzo_string_appendm_text("Newsgroups: ", poutput));
  lcrzo_epr(lcrzo_string_appendm_text(newsgroup, poutput));
  lcrzo_epr(lcrzo_string_appendm_text("\r\n", poutput));
  return(LCRZO_ERR_OK);
}

/*-------------------------------------------------------------*/
int shared_nntp_initm_headerbody_usascii(lcrzo_conststring from,
					 lcrzo_conststring replyto,
					 lcrzo_conststring newsgroup,
					 lcrzo_conststring subject, 
					 lcrzo_constdata bodydata,
					 lcrzo_int32 bodydatasize,
					 lcrzo_data *pdataout,
					 lcrzo_int32 *pdataoutsize)
{ lcrzo_int32 datasize;
  lcrzo_string header;

  /* generate the header */
  lcrzo_epr(lcrzo_string_initm_text("", &header));
  lcrzo_epr(shared_smtp_appendm_line_msgid(&header));
  lcrzo_epr(shared_smtp_appendm_line_date(&header));
  lcrzo_epr(shared_smtp_appendm_line_from(from, from, &header));
  if ( strcmp(from, replyto) )
  { lcrzo_epr(shared_smtp_appendm_line_replyto(replyto, replyto, &header));
  }
  lcrzo_epr(shared_smtp_appendm_line_organization(&header));
  lcrzo_epr(shared_smtp_appendm_line_xmailer(&header));
  lcrzo_epr(shared_smtp_appendm_line_xacclang(&header));
  lcrzo_epr(shared_nntp_appendm_line_newsgroups(newsgroup, &header));
  lcrzo_epr(shared_smtp_appendm_line_subject(subject, &header));
  lcrzo_epr(shared_smtp_appendm_line_empty(&header));

  /* append the data */
  lcrzo_epr(lcrzo_data_initm_text(header, pdataout, &datasize));
  lcrzo_epr(lcrzo_string_free(header));
  lcrzo_epr(lcrzo_data_appendm_data(bodydata, bodydatasize,
				    datasize, pdataout, &datasize));

  /* end */
  if (pdataoutsize!=NULL) *pdataoutsize=datasize;
  return(LCRZO_ERR_OK);
}

/*-------------------------------------------------------------*/
int shared_nntp_initm_headerbody_iso8859(lcrzo_conststring from,
					 lcrzo_conststring replyto,
					 lcrzo_conststring newsgroup,
					 lcrzo_conststring subject, 
					 lcrzo_constdata bodydata,
					 lcrzo_int32 bodydatasize,
					 lcrzo_data *pdataout,
					 lcrzo_int32 *pdataoutsize)
{ lcrzo_int32 datasize;
  lcrzo_string header;

  /* generate the header */
  lcrzo_epr(lcrzo_string_initm_text("", &header));
  lcrzo_epr(shared_smtp_appendm_line_msgid(&header));
  lcrzo_epr(shared_smtp_appendm_line_date(&header));
  lcrzo_epr(shared_smtp_appendm_line_from(from, from, &header));
  if ( strcmp(from, replyto) )
  { lcrzo_epr(shared_smtp_appendm_line_replyto(replyto, replyto, &header));
  }
  lcrzo_epr(shared_smtp_appendm_line_organization(&header));
  lcrzo_epr(shared_smtp_appendm_line_xmailer(&header));
  lcrzo_epr(shared_smtp_appendm_line_xacclang(&header));
  lcrzo_epr(shared_smtp_appendm_line_mimeversion(&header));
  lcrzo_epr(shared_nntp_appendm_line_newsgroups(newsgroup, &header));
  lcrzo_epr(shared_smtp_appendm_line_subject(subject, &header));
  lcrzo_epr(shared_smtp_appendm_line_conttype_textplainiso8859(NULL, &header));
  lcrzo_epr(shared_smtp_appendm_line_cte_8bit(&header));
  lcrzo_epr(shared_smtp_appendm_line_empty(&header));

  /* append the data */
  lcrzo_epr(lcrzo_data_initm_text(header, pdataout, &datasize));
  lcrzo_epr(lcrzo_string_free(header));
  lcrzo_epr(lcrzo_data_appendm_data(bodydata, bodydatasize,
				    datasize, pdataout, &datasize));

  /* end */
  if (pdataoutsize!=NULL) *pdataoutsize=datasize;
  return(LCRZO_ERR_OK);
}


/*-------------------------------------------------------------*/
int shared_nntp_initm_multipart_header(lcrzo_conststring from,
				       lcrzo_conststring replyto,
				       lcrzo_conststring newsgroup,
				       lcrzo_conststring subject, 
				       lcrzo_string *pboundary,
				       lcrzo_data *pdataout,
				       lcrzo_int32 *pdataoutsize)
{ lcrzo_string header;

  lcrzo_epr(shared_smtp_generateboundary(pboundary));

  /* generate the header */
  lcrzo_epr(lcrzo_string_initm_text("", &header));
  lcrzo_epr(shared_smtp_appendm_line_msgid(&header));
  lcrzo_epr(shared_smtp_appendm_line_date(&header));
  lcrzo_epr(shared_smtp_appendm_line_from(from, from, &header));
  if ( strcmp(from, replyto) )
  { lcrzo_epr(shared_smtp_appendm_line_replyto(replyto, replyto, &header));
  }
  lcrzo_epr(shared_smtp_appendm_line_organization(&header));
  lcrzo_epr(shared_smtp_appendm_line_xmailer(&header));
  lcrzo_epr(shared_smtp_appendm_line_xacclang(&header));
  lcrzo_epr(shared_smtp_appendm_line_mimeversion(&header));
  lcrzo_epr(shared_nntp_appendm_line_newsgroups(newsgroup, &header));
  lcrzo_epr(shared_smtp_appendm_line_subject(subject, &header));
  lcrzo_epr(shared_smtp_appendm_line_conttype_mpmixed(*pboundary, &header));
  lcrzo_epr(shared_smtp_appendm_line_empty(&header));

  /* append the data */
  lcrzo_epr(lcrzo_string_appendm_text(
          "This is a multi-part message in MIME format.\r\n", &header));

  /* generate output data */
  lcrzo_epr(lcrzo_data_initm_text(header, pdataout, pdataoutsize));
  lcrzo_epr(lcrzo_string_free(header));

  return(LCRZO_ERR_OK);
}


/*-------------------------------------------------------------*/
int shared_nntp_sendmsg(lcrzo_ipl serverip, 
			lcrzo_uint16 serverport,
			lcrzo_conststring newsgroup, 
			lcrzo_constdata headerbody,
			lcrzo_int32 headerbodysize)
{ lcrzo_sock sock;
  lcrzo_data data;
  lcrzo_int32 datasize;

  /* initialize the client */
  lcrzo_epr(lcrzo_sock_tcpcli_real(serverip, serverport, &sock));

  /* first, we read the banner */
  lcrzo_epr(shared_misc_verifyresponse1(&sock, 200));

  /* then, we send a 'MODE READER' */
  lcrzo_epr(lcrzo_data_initm_text("MODE READER\r\n", &data, &datasize));
  lcrzo_epr(lcrzo_sock_write(&sock, data, datasize));
  lcrzo_data_free(data);
  lcrzo_epr(shared_misc_verifyresponse1(&sock, 200));

  /* then, we send a 'GROUP xx' */
  lcrzo_epr(lcrzo_data_initm_text("GROUP ", &data, &datasize));
  lcrzo_epr(lcrzo_data_appendm_text(newsgroup, datasize, &data, &datasize));
  lcrzo_epr(lcrzo_data_appendm_text("\r\n", datasize, &data, &datasize));
  lcrzo_epr(lcrzo_sock_write(&sock, data, datasize));
  lcrzo_data_free(data);
  lcrzo_epr(shared_misc_verifyresponse1(&sock, 211));

  /* then, we send a 'POST' */
  lcrzo_epr(lcrzo_data_initm_text("POST\r\n", &data, &datasize));
  lcrzo_epr(lcrzo_sock_write(&sock, data, datasize));
  lcrzo_data_free(data);
  lcrzo_epr(shared_misc_verifyresponse1(&sock, 340));

  /* then, we send data */
  lcrzo_epr(lcrzo_sock_write(&sock, headerbody, headerbodysize));
  lcrzo_epr(lcrzo_data_initm_text(".\r\n", &data, &datasize));
  lcrzo_epr(lcrzo_sock_write(&sock, data, datasize));
  lcrzo_data_free(data);
  lcrzo_epr(shared_misc_verifyresponse1(&sock, 240));

  /* then, we send a 'QUIT' */
  lcrzo_epr(lcrzo_data_initm_text("QUIT\r\n", &data, &datasize));
  lcrzo_epr(lcrzo_sock_write(&sock, data, datasize));
  lcrzo_data_free(data);
  lcrzo_epr(shared_misc_verifyresponse1(&sock, 205));

  /* close the socket */
  lcrzo_epr(lcrzo_sock_close(&sock));
  return(LCRZO_ERR_OK);
}


/*-------------------------------------------------------------*/
int shared_nntp_initm_headerbody_file(lcrzo_conststring from,
				      lcrzo_conststring replyto,
				      lcrzo_conststring newsgroup,
				      lcrzo_conststring subject, 
				      lcrzo_conststring bodyfilename,
				      lcrzo_data *pdata,
				      lcrzo_int32 *pdatasize)
{ lcrzo_data bodydata;
  lcrzo_int32 bodydatasize;
  lcrzo_string thesubject;

  lcrzo_epr(shared_smtp_initm_subjectdata_file(subject, bodyfilename,
					       &thesubject,
					       &bodydata, &bodydatasize));

  lcrzo_epr(shared_nntp_initm_headerbody_iso8859(from, replyto, newsgroup, 
						 thesubject, 
						 bodydata, bodydatasize,
						 pdata, pdatasize));
  lcrzo_data_free(bodydata);
  lcrzo_string_free(thesubject);
  return(LCRZO_ERR_OK);
}

/*-------------------------------------------------------------*/
int shared_nntp_initm_multipart_2files(lcrzo_conststring from,
				       lcrzo_conststring replyto,
				       lcrzo_conststring newsgroup,
				       lcrzo_conststring subject, 
				       lcrzo_conststring bodyfilename,
				       lcrzo_conststring attachementfilename,
				       lcrzo_data *pdata,
				       lcrzo_int32 *pdatasize)
{ lcrzo_data bodydata;
  lcrzo_int32 bodydatasize, datasize;
  lcrzo_string thesubject, boundary, tmp;

  lcrzo_epr(shared_smtp_initm_subjectdata_file(subject, bodyfilename,
					       &thesubject,
					       &bodydata, &bodydatasize));

  /* header */
  lcrzo_epr(shared_nntp_initm_multipart_header(from, replyto, newsgroup,
					       thesubject, 
					       &boundary, pdata, &datasize));
  lcrzo_string_free(thesubject);

  /* first file */
  lcrzo_epr(lcrzo_filename_lastfile(bodyfilename, &tmp));
  lcrzo_epr(shared_smtp_appendm_multipart_iso8859(boundary, tmp, tmp,
						  bodydata, bodydatasize,
						  datasize, pdata, &datasize));
  lcrzo_string_free(tmp);
  lcrzo_data_free(bodydata);

  /* second file */
  lcrzo_epr(lcrzo_file_read(attachementfilename, &bodydata, &bodydatasize));
  lcrzo_epr(lcrzo_filename_lastfile(attachementfilename, &tmp));
  lcrzo_epr(shared_smtp_appendm_multipart_base64(boundary, tmp, tmp,
						 bodydata, bodydatasize,
						 datasize, pdata, &datasize));
  lcrzo_string_free(tmp);
  lcrzo_data_free(bodydata);

  /* end */
  lcrzo_epr(shared_smtp_appendm_multipart_end(boundary, 
					      datasize, pdata, &datasize));

  if (pdatasize!=NULL) *pdatasize=datasize;
  return(LCRZO_ERR_OK);
}

/*-------------------------------------------------------------*/
int shared_nntp_sendmsg_hb_file(lcrzo_ipl serverip, 
				lcrzo_uint16 serverport,
				lcrzo_conststring from, 
				lcrzo_conststring replyto, 
				lcrzo_conststring newsgroup,
				lcrzo_conststring subject, 
				lcrzo_conststring bodyfilename)
{ lcrzo_data headerbody;
  lcrzo_int32 headerbodysize;

  lcrzo_epr(shared_nntp_initm_headerbody_file(from, replyto, newsgroup,
					      subject,
					      bodyfilename, &headerbody,
					      &headerbodysize));

  lcrzo_epr(shared_nntp_sendmsg(serverip, serverport, newsgroup,
				headerbody, headerbodysize));
  lcrzo_data_free(headerbody);
  return(LCRZO_ERR_OK);
}

/*-------------------------------------------------------------*/
int shared_nntp_sendmsg_mp_2files(lcrzo_ipl serverip, 
				  lcrzo_uint16 serverport,
				  lcrzo_conststring from, 
				  lcrzo_conststring replyto, 
				  lcrzo_conststring newsgroup,
				  lcrzo_conststring subject, 
				  lcrzo_conststring bodyfilename,
				  lcrzo_conststring attachementfilename)
{ lcrzo_data headerbody;
  lcrzo_int32 headerbodysize;

  lcrzo_epr(shared_nntp_initm_multipart_2files(from, replyto, newsgroup, 
					       subject, bodyfilename, 
					       attachementfilename,
					       &headerbody,
					       &headerbodysize));

  lcrzo_epr(shared_nntp_sendmsg(serverip, serverport, newsgroup,
				headerbody, headerbodysize));
  lcrzo_data_free(headerbody);
  return(LCRZO_ERR_OK);
}

/*-------------------------------------------------------------*/
void shared_nntp_list_newsgroups_free(void *pelem);
void shared_nntp_list_newsgroups_free(void *pelem)
{ shared_nntp_list_newsgroups *psnln;
  psnln=(shared_nntp_list_newsgroups *)pelem;
  lcrzo_data_free(psnln->name);
}

/*-------------------------------------------------------------*/
int shared_nntp_list_init_newsgroups(lcrzo_ipl serverip, 
				     lcrzo_uint16 serverport,
				     lcrzo_list *plist)
{ lcrzo_sock sock;
  lcrzo_regexp re;
  lcrzo_data data;
  lcrzo_int32 datasize;
  shared_nntp_list_newsgroups snln;
  lcrzo_string tmp;
  int ret;

  /* initialize the client */
  lcrzo_epr(lcrzo_sock_tcpcli_real(serverip, serverport, &sock));

  /* first, we read the banner */
  lcrzo_epr(shared_misc_verifyresponse1(&sock, 200));

  /* then, we send a 'MODE READER' */
  lcrzo_epr(lcrzo_data_initm_text("MODE READER\r\n", &data, &datasize));
  lcrzo_epr(lcrzo_sock_write(&sock, data, datasize));
  lcrzo_data_free(data);
  lcrzo_epr(shared_misc_verifyresponse1(&sock, 200));

  /* then, we send a 'LIST' */
  lcrzo_epr(lcrzo_data_initm_text("LIST\r\n", &data, &datasize));
  lcrzo_epr(lcrzo_sock_write(&sock, data, datasize));
  lcrzo_data_free(data);
  lcrzo_epr(shared_misc_verifyresponse1(&sock, 215));

  /* initialize the chained list */
  lcrzo_epr(lcrzo_list_init(plist, sizeof(snln),
			    &shared_nntp_list_newsgroups_free));

  while (LCRZO_TRUE)
  { /* read line */
    lcrzo_epr(lcrzo_sock_readm_line(&sock, LCRZO_TRUE, &data, &datasize));
    /*lcrzo_epr(lcrzo_data_print(data, datasize, LCRZO_PRINTTYPE_DUMP));*/
    if ( datasize==1 && data[0]=='.' )
    { lcrzo_data_free(data);
      break;
    }
    /* analyze line */
    data[datasize]='\0';
    ret=lcrzo_string_search_regexp(data, LCRZO_FALSE, +1, -1,
				   "^(.+) +(.+) +(.+) +(.+)$",re);
    if ( ret!=LCRZO_ERR_OK )
    { lcrzo_data_free(data);
      continue;
    }
    lcrzo_epr(lcrzo_string_initm_range(data, re[1].startindex_positive,
				       re[1].endindex_positive, &(snln.name)));
    lcrzo_epr(lcrzo_string_initm_range(data, re[2].startindex_positive,
				       re[2].endindex_positive, &tmp));
    lcrzo_uint32_init_text(tmp, &(snln.numberlast));
    lcrzo_string_free(tmp);
    lcrzo_epr(lcrzo_string_initm_range(data, re[3].startindex_positive,
				       re[3].endindex_positive, &tmp));
    lcrzo_uint32_init_text(tmp, &(snln.numberfirst));
    lcrzo_string_free(tmp);
    lcrzo_epr(lcrzo_string_initm_range(data, re[4].startindex_positive,
				       re[4].endindex_positive, &tmp));
    if (tmp[0]=='y') snln.canpost=LCRZO_TRUE;
    else snln.canpost=LCRZO_FALSE;
    lcrzo_string_free(tmp);
    /* add the elem */
    lcrzo_epr(lcrzo_list_add_last(plist, &snln));
    /* continue */
    lcrzo_data_free(data);
  }

  /* then, we send a 'QUIT' */
  lcrzo_epr(lcrzo_data_initm_text("QUIT\r\n", &data, &datasize));
  lcrzo_epr(lcrzo_sock_write(&sock, data, datasize));
  lcrzo_data_free(data);
  lcrzo_epr(shared_misc_verifyresponse1(&sock, 205));

  /* close the socket */
  lcrzo_epr(lcrzo_sock_close(&sock));
  return(LCRZO_ERR_OK);
}

/*-------------------------------------------------------------*/
int shared_nntp_recvmsg(lcrzo_ipl serverip, 
			lcrzo_uint16 serverport,
			lcrzo_conststring newsgroup, 
			lcrzo_uint32 messagenumber,
			lcrzo_data *pdataout,
			lcrzo_int32 *pdataoutsize)
{ lcrzo_sock sock;
  lcrzo_data data;
  lcrzo_int32 datasize, dataoutsize;

  /* initialize the client */
  lcrzo_epr(lcrzo_sock_tcpcli_real(serverip, serverport, &sock));

  /* first, we read the banner */
  lcrzo_epr(shared_misc_verifyresponse1(&sock, 200));

  /* then, we send a 'MODE READER' */
  lcrzo_epr(lcrzo_data_initm_text("MODE READER\r\n", &data, &datasize));
  lcrzo_epr(lcrzo_sock_write(&sock, data, datasize));
  lcrzo_data_free(data);
  lcrzo_epr(shared_misc_verifyresponse1(&sock, 200));

  /* then, we send a 'GROUP xx' */
  lcrzo_epr(lcrzo_data_initm_text("GROUP ", &data, &datasize));
  lcrzo_epr(lcrzo_data_appendm_text(newsgroup, datasize, &data, &datasize));
  lcrzo_epr(lcrzo_data_appendm_text("\r\n", datasize, &data, &datasize));
  lcrzo_epr(lcrzo_sock_write(&sock, data, datasize));
  lcrzo_data_free(data);
  lcrzo_epr(shared_misc_verifyresponse1(&sock, 211));

  /* then, we send a 'ARTICLE number' */
  lcrzo_epr(lcrzo_data_initm_text("ARTICLE ", &data, &datasize));
  lcrzo_epr(lcrzo_data_appendm_int(messagenumber, "%d", 
				   datasize, &data, &datasize));
  lcrzo_epr(lcrzo_data_appendm_text("\r\n", datasize, &data, &datasize));
  lcrzo_epr(lcrzo_sock_write(&sock, data, datasize));
  lcrzo_data_free(data);
  lcrzo_epr(shared_misc_verifyresponse1(&sock, 220));

  lcrzo_epr(lcrzo_data_initm_text("", pdataout, &dataoutsize));
  while (LCRZO_TRUE)
  { /* read line */
    lcrzo_epr(lcrzo_sock_readm_line(&sock, LCRZO_TRUE, &data, &datasize));
    /*lcrzo_epr(lcrzo_data_print(data, datasize, LCRZO_PRINTTYPE_DUMP));*/
    if ( datasize==1 && data[0]=='.' )
    { lcrzo_data_free(data);
      break;
    }
    /* append line */
    lcrzo_epr(lcrzo_data_appendm_data(data, datasize, dataoutsize, 
				      pdataout, &dataoutsize));
    lcrzo_epr(lcrzo_data_appendm_text("\n", dataoutsize,
				      pdataout, &dataoutsize));
   /* continue */
    lcrzo_data_free(data);
  }
  if (pdataoutsize!=NULL) *pdataoutsize=dataoutsize;

  /* then, we send a 'QUIT' */
  lcrzo_epr(lcrzo_data_initm_text("QUIT\r\n", &data, &datasize));
  lcrzo_epr(lcrzo_sock_write(&sock, data, datasize));
  lcrzo_data_free(data);
  lcrzo_epr(shared_misc_verifyresponse1(&sock, 205));

  /* close the socket */
  lcrzo_epr(lcrzo_sock_close(&sock));
  return(LCRZO_ERR_OK);
}


