#include <lcrzo.h>
#include "../lcrzoex.h"
#include "shared_telnet.h"
#include "shared_error.h"


/*-------------------------------------------------------------*/
int shared_priv_telnet_writeuser(lcrzo_telnet *ptelnet,
				 lcrzo_sock *psock);
int shared_priv_telnet_writeuser(lcrzo_telnet *ptelnet,
				 lcrzo_sock *psock)
{ lcrzo_data data, sentdata;
  lcrzo_int32 datasize, sentdatasize;
  int ret, i;

  /* read */
  ret=lcrzo_ipc_readm(&(ptelnet->write), LCRZO_FALSE, &data, &datasize);
  if ( ret==LCRZO_ERR_OKTEMPDATAEND ) return(LCRZO_ERR_OK);
  else if (ret!=LCRZO_ERR_OK) return(ret);

  /* convert 0xFF to 0xFF0xFF */
  lcrzo_epr(lcrzo_data_alloc(datasize, &sentdata));
  for ( i=0,sentdatasize=0 ; i<datasize ; i++,sentdatasize++ )
  { if ( data[i]!=0xFF )
    { sentdata[sentdatasize]=data[i];
    }
    else
    { sentdata[sentdatasize++]=0xFF;
      lcrzo_epr(lcrzo_data_realloc(datasize+(sentdatasize-i), &sentdata));
      sentdata[sentdatasize]=0xFF;
    }
  }
  lcrzo_data_free(data);

  /* send data */
  lcrzo_epr(lcrzo_sock_write(psock, sentdata, sentdatasize));
  lcrzo_data_free(sentdata);

  return(LCRZO_ERR_OK);
}

/*-------------------------------------------------------------*/
int shared_priv_telnet_readmoreserverdata(lcrzo_sock *psock, 
					  lcrzo_data *pdata, 
					  lcrzo_int32 *pdatasize, 
					  lcrzo_int32 *pi,
					  lcrzo_int32 neededsize);
int shared_priv_telnet_readmoreserverdata(lcrzo_sock *psock, 
					  lcrzo_data *pdata, 
					  lcrzo_int32 *pdatasize, 
					  lcrzo_int32 *pi,
					  lcrzo_int32 neededsize)
{ lcrzo_data readdata;
  lcrzo_int32 readdatasize;

  while (LCRZO_TRUE)
  { if ( *pdatasize - *pi > neededsize ) return(LCRZO_ERR_OK);
    lcrzo_epr(lcrzo_sock_readm(psock, LCRZO_TRUE, &readdata, &readdatasize));
    lcrzo_epr(lcrzo_data_appendm_data(readdata, readdatasize,
				      *pdatasize, pdata, pdatasize));
    lcrzo_data_free(readdata);
  }
  return(LCRZO_ERR_IEINTERNALERROR); 
}

/*-------------------------------------------------------------*/
int shared_priv_telnet_option_answer(lcrzo_telnet *ptelnet,
				     lcrzo_sock *psock, 
				     lcrzo_data *pdata, 
				     lcrzo_int32 *pdatasize, 
				     lcrzo_int32 *pi);
int shared_priv_telnet_option_answer(lcrzo_telnet *ptelnet,
				     lcrzo_sock *psock, 
				     lcrzo_data *pdata, 
				     lcrzo_int32 *pdatasize, 
				     lcrzo_int32 *pi)
{ lcrzo_data sentdata;
  lcrzo_int32 sentdatasize;

  /* this function should be called like this */
  if ( *pdatasize - *pi < 1 ) return(LCRZO_ERR_IEINTERNALERROR);
  if ( (*pdata)[*pi]!=0xFF ) return(LCRZO_ERR_IEINTERNALERROR);

  /* eventually read more data from server */
  lcrzo_epr(shared_priv_telnet_readmoreserverdata(psock, pdata, pdatasize,
						  pi, 1));

  (*pi)++;
  switch((*pdata)[*pi]) {
  case 0xF1: /* NOP*/
    break;
  case 0xFD: /* DO */
  case 0xFE: /* DON'T */
    /* eventually read more data from server */
    lcrzo_epr(shared_priv_telnet_readmoreserverdata(psock, pdata, pdatasize,
						    pi, 1));
    /* answer WON'T */
    sentdatasize=3;
    lcrzo_epr(lcrzo_data_alloc(sentdatasize, &sentdata));
    (*pi)++;
    sentdata[0]=0xFF; sentdata[1]=0xFC; sentdata[2]=(*pdata)[*pi];
    lcrzo_epr(lcrzo_sock_write(psock, sentdata, sentdatasize));
    lcrzo_data_free(sentdata);
    break;
  case 0xFB: /* WILL */
  case 0xFC: /* WON'T */
    /* eventually read more data from server */
    lcrzo_epr(shared_priv_telnet_readmoreserverdata(psock, pdata, pdatasize,
						    pi, 1));
    /* answer DON'T */
    sentdatasize=3;
    lcrzo_epr(lcrzo_data_alloc(sentdatasize, &sentdata));
    (*pi)++;
    sentdata[0]=0xFF; sentdata[1]=0xFE; sentdata[2]=(*pdata)[*pi];
    lcrzo_epr(lcrzo_sock_write(psock, sentdata, sentdatasize));
    lcrzo_data_free(sentdata);
    break;
  case 0xFF: /* char 0xFF */
    sentdatasize=1;
    lcrzo_epr(lcrzo_data_alloc(sentdatasize, &sentdata));
    sentdata[0]=0xFF;
    lcrzo_epr(lcrzo_ipc_write(&(ptelnet->read), sentdata, sentdatasize));
    lcrzo_data_free(sentdata);
    break;
  default:
    printf("Warning : could not decode 0xFF%02X\n", (*pdata)[*pi]);
    return(123456);
    break;
  }

  return(LCRZO_ERR_OK);
}

/*-------------------------------------------------------------*/
int shared_priv_telnet_readserver(lcrzo_telnet *ptelnet,
				  lcrzo_sock *psock);
int shared_priv_telnet_readserver(lcrzo_telnet *ptelnet,
				  lcrzo_sock *psock)
{ lcrzo_data data, sentdata;
  lcrzo_int32 datasize, sentdatasize, i;
  lcrzo_bool sentdatainit;
  int ret;

  /* read server data*/
  ret=lcrzo_sock_readm(psock, LCRZO_FALSE, &data, &datasize);
  if ( ret==LCRZO_ERR_OKTEMPDATAEND ) return(LCRZO_ERR_OK);
  else if (ret!=LCRZO_ERR_OK) return(ret);

  /* analyze its data, answer negatively to requests, and sent data to user */
  sentdatainit=LCRZO_FALSE;
  sentdatasize=0;
  for ( i=0 ; i<datasize ; i++ )
  { /* data for user */
    if ( data[i]!=0xFF )
    { if (!sentdatainit)
      { lcrzo_epr(lcrzo_data_alloc(datasize, &sentdata));
        sentdatasize=0;
        sentdatainit=LCRZO_TRUE;
      }
      sentdata[sentdatasize++]=data[i];
    }
    /* control data */
    else
    { /* first send data for user */
      if (sentdatainit)
      { lcrzo_epr(lcrzo_ipc_write(&(ptelnet->read), sentdata, sentdatasize));
        lcrzo_data_free(sentdata);
	sentdatainit=LCRZO_FALSE;
      }
      /* decode this option */
      lcrzo_epr(shared_priv_telnet_option_answer(ptelnet, psock, 
						 &data, &datasize, &i));
    }
    /* we have suffeciently data */
    if (sentdatainit && sentdatasize>1000)
    { lcrzo_epr(lcrzo_ipc_write(&(ptelnet->read), sentdata, sentdatasize));
      lcrzo_data_free(sentdata);
      sentdatainit=LCRZO_FALSE;
    }
  }
  lcrzo_data_free(data);

  /* eventually send data for user */
  if (sentdatainit)
  { lcrzo_epr(lcrzo_ipc_write(&(ptelnet->read), sentdata, sentdatasize));
    lcrzo_data_free(sentdata);
  }

  return(LCRZO_ERR_OK);
}

/*-------------------------------------------------------------*/
int shared_priv_telnet_child(lcrzo_ipl serveripl, 
			     lcrzo_uint16 serverport,
			     lcrzo_telnet *ptelnet);
int shared_priv_telnet_child(lcrzo_ipl serveripl, 
			     lcrzo_uint16 serverport,
			     lcrzo_telnet *ptelnet)
{ lcrzo_sock sock;
  lcrzo_bool canreadserver, canwriteuser;
  int fdreadserver, fdwriteuser, ret;

  /* init the tcp client */
  lcrzo_epr(lcrzo_sock_tcpcli_real(serveripl, serverport, &sock));

  /* close unneeded IPCs */
  lcrzo_epr(lcrzo_ipc_close_read(&(ptelnet->read)));
  lcrzo_epr(lcrzo_ipc_close_write(&(ptelnet->write)));
  lcrzo_epr(lcrzo_ipc_close_write(&(ptelnet->controle)));

  /* end of negociation */
  lcrzo_epr(lcrzo_ipc_write_uint32(&(ptelnet->read), LCRZO_ERR_OK));

  /* get file descriptors for select */
  lcrzo_epr(lcrzo_sock_get_fds(&sock, &fdreadserver, NULL));
  lcrzo_epr(lcrzo_ipc_get_fds(&(ptelnet->write), &fdwriteuser, NULL));
  while(LCRZO_TRUE)
  { lcrzo_epr(lcrzo_fd_select_read2(fdreadserver, fdwriteuser, NULL,
				    &canreadserver, &canwriteuser));
  /*printf("%d_%d    %d_%d\n", fdreadserver, canreadserver, 
    fdwriteuser, canwriteuser);*/
    if ( canreadserver )
    { /* server has data */
      ret=shared_priv_telnet_readserver(ptelnet, &sock);
      if ( ret==LCRZO_ERR_OKDATAEND )
      { /* server has closed connection : do not ask it anymore */
        fdreadserver=-1;
      }
    }
    if ( canwriteuser )
    { /* user wants to write */
      lcrzo_epr(shared_priv_telnet_writeuser(ptelnet, &sock));
    }
  }

  return(LCRZO_ERR_OK);
}

/*-------------------------------------------------------------*/
int shared_priv_telnet_ipcreadtime(lcrzo_ipc *pipc,
				   lcrzo_uint32 readtimeout,
				   lcrzo_data *preaddata,
				   lcrzo_int32 *preaddatasize);
int shared_priv_telnet_ipcreadtime(lcrzo_ipc *pipc,
				   lcrzo_uint32 readtimeout,
				   lcrzo_data *preaddata,
				   lcrzo_int32 *preaddatasize)
{ int fdread;
  lcrzo_time timeout;
  lcrzo_data data, readdata;
  lcrzo_int32 datasize, readdatasize;
  lcrzo_bool canread;

  lcrzo_epr(lcrzo_time_init_usec(readtimeout, &timeout));
  lcrzo_epr(lcrzo_data_initm_text("", &readdata, &readdatasize));
  lcrzo_epr(lcrzo_ipc_get_fds(pipc, &fdread, NULL));

  /* read data */
  while (LCRZO_TRUE)
  { lcrzo_epr(lcrzo_fd_select_read1(fdread, &timeout, &canread));
    if ( ! canread ) break;
    lcrzo_epr(lcrzo_ipc_readm(pipc, LCRZO_TRUE, &data, &datasize));
    lcrzo_epr(lcrzo_data_appendm_data(data, datasize, readdatasize, 
				      &readdata, &readdatasize));
    lcrzo_data_free(data);
  }

  /* set output data */
  if (preaddatasize!=NULL) *preaddatasize=readdatasize;
  if (preaddata==NULL)
  { lcrzo_data_free(readdata);
  }
  else 
  { *preaddata=readdata;
  }
  return(LCRZO_ERR_OK);
}

/*-------------------------------------------------------------*/
/*-------------------------------------------------------------*/
/*-------------------------------------------------------------*/
int shared_telnet_init(lcrzo_ipl serveripl, 
		       lcrzo_uint16 serverport,
		       lcrzo_telnet *ptelnet)
{ lcrzo_uint32 ret;
  int retchild;

  /* open two IPCs */
  lcrzo_epr(lcrzo_ipc_init(&(ptelnet->read)));
  lcrzo_epr(lcrzo_ipc_init(&(ptelnet->write)));
  lcrzo_epr(lcrzo_ipc_init(&(ptelnet->controle)));

  /* create the child */
  lcrzo_epr(lcrzo_process_fork(&(ptelnet->pid)));
  if (ptelnet->pid==0)
  { retchild=shared_priv_telnet_child(serveripl, serverport, ptelnet);
    /*printf("Child exit %d\n", retchild);*/
    lcrzo_process_exit(retchild);
    /*never reatched*/
  }
  else
  { /* close unneeded IPCs */
    lcrzo_epr(lcrzo_ipc_close_write(&(ptelnet->read)));
    lcrzo_epr(lcrzo_ipc_close_read(&(ptelnet->write)));
    lcrzo_epr(lcrzo_ipc_close_read(&(ptelnet->controle)));
  }

  /* wait the end of negociation */
  lcrzo_epr(lcrzo_ipc_read_uint32(&(ptelnet->read), LCRZO_TRUE, &ret));

  return((int)ret);
}

/*-------------------------------------------------------------*/
int shared_telnet_close(lcrzo_telnet *ptelnet)
{
  /* kill child */
  lcrzo_epr(lcrzo_process_killwait(ptelnet->pid));

  /* close IPCs */
  lcrzo_epr(lcrzo_ipc_close(&(ptelnet->read)));
  lcrzo_epr(lcrzo_ipc_close(&(ptelnet->write)));
  lcrzo_epr(lcrzo_ipc_close(&(ptelnet->controle)));

  return(LCRZO_ERR_OK);
}

/*-------------------------------------------------------------*/
int shared_telnet_login(lcrzo_telnet *ptelnet,
			lcrzo_conststring login,
			lcrzo_conststring password)
{ lcrzo_data data, banner;
  lcrzo_int32 datasize, bannersize;
  lcrzo_regexp re;
  lcrzo_bool regcase;
  int ret;

  /* read till "login: "*/
  lcrzo_epr(lcrzo_global_get_data_regexp_casesensitive(&regcase));
  lcrzo_epr(lcrzo_global_set_data_regexp_casesensitive(LCRZO_FALSE));
  lcrzo_epr(lcrzo_data_initm_text("", &banner, &bannersize));
  while (LCRZO_TRUE)
  { lcrzo_epr(lcrzo_ipc_readm(&(ptelnet->read), LCRZO_TRUE, &data, &datasize));
    lcrzo_epr(lcrzo_data_appendm_data(data, datasize, bannersize,
				      &banner, &bannersize));
    lcrzo_epr(lcrzo_data_free(data));
    ret=lcrzo_data_search_regexp(banner, bannersize, LCRZO_FALSE, +1, -1,
                                 "login[ \t]*:[ \t]*$", re);
    if ( ret==LCRZO_ERR_OK ) break;
  }
  lcrzo_epr(lcrzo_data_free(banner));

  /* send login */
  lcrzo_epr(lcrzo_data_initm_text(login, &data, &datasize));
  lcrzo_epr(lcrzo_data_appendm_text("\r\n", datasize, &data, &datasize));
  lcrzo_epr(lcrzo_ipc_write(&(ptelnet->write), data, datasize));
  lcrzo_epr(lcrzo_data_free(data));

  /* read till "word: "*/
  lcrzo_epr(lcrzo_data_initm_text("", &banner, &bannersize));
  while (LCRZO_TRUE)
  { lcrzo_epr(lcrzo_ipc_readm(&(ptelnet->read), LCRZO_TRUE, &data, &datasize));
    lcrzo_epr(lcrzo_data_appendm_data(data, datasize, bannersize,
				      &banner, &bannersize));
    lcrzo_epr(lcrzo_data_free(data));
    ret=lcrzo_data_search_regexp(banner, bannersize, LCRZO_FALSE, +1, -1,
                                 "word[ \t]*:[ \t]*$", re);
    if ( ret==LCRZO_ERR_OK ) break;
  }
  lcrzo_epr(lcrzo_data_free(banner));

  /* send password */
  lcrzo_epr(lcrzo_data_initm_text(password, &data, &datasize));
  lcrzo_epr(lcrzo_data_appendm_text("\r\n", datasize, &data, &datasize));
  lcrzo_epr(lcrzo_ipc_write(&(ptelnet->write), data, datasize));
  lcrzo_epr(lcrzo_data_free(data));
  /* read \r\n sent by server */
  lcrzo_epr(lcrzo_ipc_readm_line(&(ptelnet->read), LCRZO_TRUE, NULL, NULL));
    
  /* read anwser */
  lcrzo_epr(lcrzo_ipc_readm_line(&(ptelnet->read), LCRZO_TRUE, 
				 &banner, &bannersize));
  ret=lcrzo_data_search_regexp(banner, bannersize, LCRZO_FALSE, +1, -1,
			       "^Login incorrect", re);
  lcrzo_epr(lcrzo_data_free(banner));
  lcrzo_epr(lcrzo_global_set_data_regexp_casesensitive(regcase));

  if ( ret==LCRZO_ERR_OK )
  { /* automatically exit telnet session */
    lcrzo_epr(shared_telnet_close(ptelnet));
    return(LCRZOEX_ERR_TELNET_BADLOGINPASS);
  }
 
  /* read all banners */
  lcrzo_epr(shared_priv_telnet_ipcreadtime(&(ptelnet->read), 1300000,
					   NULL, NULL));

  return(LCRZO_ERR_OK);
}

/*-------------------------------------------------------------*/
int shared_telnet_exec(lcrzo_telnet *ptelnet,
		       lcrzo_conststring command,
		       lcrzo_uint32 readtimeout,
		       lcrzo_data *preaddata,
		       lcrzo_int32 *preaddatasize)
{ lcrzo_data data;
  lcrzo_int32 datasize;

  /* send command */
  lcrzo_epr(lcrzo_data_initm_text(command, &data, &datasize));
  lcrzo_epr(lcrzo_data_appendm_text("\r\n", datasize, &data, &datasize));
  lcrzo_epr(lcrzo_ipc_write(&(ptelnet->write), data, datasize));
  lcrzo_epr(lcrzo_data_free(data));

  /* read result */
  lcrzo_epr(shared_priv_telnet_ipcreadtime(&(ptelnet->read), readtimeout,
					   preaddata, preaddatasize));

  return(LCRZO_ERR_OK);
}

/*-------------------------------------------------------------*/
int shared_telnet_logout(lcrzo_telnet *ptelnet)
{ lcrzo_data data;
  lcrzo_int32 datasize;
  int ret;

  /* send command */
  lcrzo_epr(lcrzo_data_initm_text("exit\r\n", &data, &datasize));
  lcrzo_epr(lcrzo_ipc_write(&(ptelnet->write), data, datasize));
  lcrzo_epr(lcrzo_data_free(data));

  /* read result */
  ret=shared_priv_telnet_ipcreadtime(&(ptelnet->read), 100000, NULL, NULL);
  if ( ret==LCRZO_ERR_OKDATAEND || ret==LCRZO_ERR_OK ) return(LCRZO_ERR_OK);
  return(ret);
}

